import json
from itertools import chain
from typing import Optional, Tuple  # pylint: disable=unused-import

from podite import pod_json, field, named_fields, Vec, Enum, Variant
from podite.types.enum import ENUM_TAG_NAME, ENUM_TAG_NAME_MAP
from podite.json import POD_OPTIONS_RENAME

from dexteritysdk.solmate.utils import camel_to_snake


def camel_case(name):
    head, *tail = name.split("_")
    return "".join(chain([head.lower()], map(str.capitalize, tail)))


@pod_json
class IdlType(Enum):
    __enum_options__ = {ENUM_TAG_NAME_MAP: camel_case}

    BOOL = None
    U8 = None
    I8 = None
    U16 = None
    I16 = None
    U32 = None
    I32 = None
    U64 = None
    I64 = None
    U128 = None
    I128 = None
    BYTES = None
    STRING = None
    PUBLIC_KEY = None
    DEFINED = Variant(field=str)

    OPTION = Variant(field="IdlType")
    COPTION = Variant(field="IdlType")
    STATIC = Variant(field="IdlType")
    VEC = Variant(field="IdlType")
    ARRAY = Variant(field="Tuple[IdlType, int]")


@pod_json
class IdlField:
    name: str
    type: IdlType

    @property
    def py_name(self):
        return camel_to_snake(self.name)


@pod_json
class EnumFields(Enum):
    NAMED = Variant(field=Vec[IdlField])
    TUPLE = Variant(field=Vec[IdlType])

    @classmethod
    def _from_dict(cls, raw):
        assert isinstance(raw, list)
        if len(raw) == 0:
            return EnumFields.NAMED([])

        if "name" in raw[0]:
            field_val = Vec[IdlField].from_dict(raw)
            return EnumFields.NAMED(field_val)
        else:
            field_val = Vec[IdlType].from_dict(raw)
            return EnumFields.TUPLE(field_val)

    @classmethod
    def _to_dict(cls, instance):
        raise NotImplementedError


@pod_json
class IdlEnumVariant:
    name: str
    fields: Optional[EnumFields] = field(default=None)


@pod_json
class IdlTypeDefinitionTy(Enum):
    __enum_options__ = {
        ENUM_TAG_NAME: "kind",
        ENUM_TAG_NAME_MAP: "lower",
    }

    STRUCT = Variant(field=named_fields(fields=Vec[IdlField]))
    ENUM = Variant(field=named_fields(variants=Vec[IdlEnumVariant]))


@pod_json
class IdlTypeDefinition:
    name: str
    type: IdlTypeDefinitionTy


@pod_json
class IdlEventField:
    name: str
    type: IdlType
    index: bool


@pod_json
class IdlEvent:
    name: str
    fields: Vec[IdlEventField] = field(default_factory=list)


@pod_json
class IdlErrorCode:
    code: int
    name: str
    msg: Optional[str] = field(default=None)


@pod_json
class IdlConst:
    name: str
    type: IdlType
    value: str


@pod_json
class IdlAccount:
    __pod_options__ = {POD_OPTIONS_RENAME: camel_case}

    name: str
    is_mut: bool
    is_signer: bool
    is_optional: bool = field(default=False)


@pod_json
class IdlAccounts:
    __enum_options__ = {ENUM_TAG_NAME_MAP: camel_case}

    name: str
    accounts: Vec["IdlAccountItem"]


@pod_json
class IdlAccountItem(Enum):
    IDL_ACCOUNT = Variant(field=IdlAccount)
    IDL_ACCOUNTS = Variant(field=IdlAccounts)

    @classmethod
    def _from_dict(cls, raw):
        assert isinstance(raw, dict)

        if "accounts" in raw:
            name = raw["name"]
            sub_accounts = Vec[IdlAccountItem].from_dict(raw["accounts"])
            field_value = IdlAccounts(name, sub_accounts)
            return IdlAccountItem.IDL_ACCOUNTS(field_value)
        else:
            field_value = IdlAccount.from_dict(raw)
            return IdlAccountItem.IDL_ACCOUNT(field_value)

    @classmethod
    def _to_dict(cls, instance):
        raise NotImplementedError


@pod_json
class IdlInstruction:
    name: str
    accounts: Vec[IdlAccountItem]
    args: Vec[IdlField]


@pod_json
class IdlState:
    struct: IdlTypeDefinition
    methods: Vec[IdlInstruction]


@pod_json
class Idl:
    version: str
    name: str
    constants: Vec[IdlConst] = field(default_factory=list)
    instructions: Vec[IdlInstruction] = field(default_factory=list)
    state: Optional[IdlState] = field(default=None)
    accounts: Vec[IdlTypeDefinition] = field(default_factory=list)
    types: Vec[IdlTypeDefinition] = field(default_factory=list)
    events: Vec[IdlEvent] = field(default_factory=list)
    errors: Vec[IdlErrorCode] = field(default_factory=list)
    metadata: Optional[object] = field(default=None)

    @staticmethod
    def from_json_file(filename):
        with open(filename, "r") as fin:
            idl_dict = json.load(fin)
            return Idl.from_dict(idl_dict)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.operation_type import OperationType
from dexteritysdk.solmate.dtypes import Usize
from dexteritysdk.utils.aob.state.base import Side
from podite import pod

# LOCK-END


# LOCK-BEGIN[class(OrderInfo)]: DON'T MODIFY
@pod
class OrderInfo:
    total_order_qty: "Fractional"
    matched_order_qty: "Fractional"
    order_side: Side
    order_price: "Fractional"
    is_combo: bool
    product_index: Usize
    operation_type: "OperationType"
    old_ask_qty_in_book: "Fractional"
    old_bid_qty_in_book: "Fractional"
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import struct

from solders.pubkey import Pubkey
from solders.instruction import AccountMeta, Instruction
from dexteritysdk.program_ids import AOB_PROGRAM_ID


CREATE_MARKET_IX_CODE = 0


def create_market_aob_ix(
    market: Pubkey,
    event_queue: Pubkey,
    bids: Pubkey,
    asks: Pubkey,
    caller_authority: Pubkey,
    callback_info_len: int,  # u64
    callback_id_len: int,  # u64
    min_base_order_size: int,  # u64
    price_bitmask: int = (1 << 64) - 1,
    cranker_reward: int = 1000,
):
    params = [
        CREATE_MARKET_IX_CODE,
        bytes(caller_authority),
        callback_info_len,
        callback_id_len,
        min_base_order_size,
        price_bitmask,
        cranker_reward,
    ]
    return Instruction(
        accounts=[
            AccountMeta(market, is_signer=False, is_writable=True),
            AccountMeta(event_queue, is_signer=False, is_writable=True),
            AccountMeta(bids, is_signer=False, is_writable=True),
            AccountMeta(asks, is_signer=False, is_writable=True),
        ],
        program_id=AOB_PROGRAM_ID,
        data=struct.pack("<B32sQQQQQ", *params),
    )



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.solmate.anchor import InstructionDiscriminant
from podite import (
    Enum,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[instruction_tag]: DON'T MODIFY
@pod
class InstructionTag(Enum[U64]):
    INITIALIZE_MARKET_PRODUCT_GROUP = InstructionDiscriminant()
    UPDATE_MARKET_PRODUCT_GROUP = InstructionDiscriminant()
    SET_ADDRESS_LOOKUP_TABLE = InstructionDiscriminant()
    INITIALIZE_MARKET_PRODUCT = InstructionDiscriminant()
    CHANGE_ORDERBOOK = InstructionDiscriminant()
    DEACTIVATE_MARKET_PRODUCT = InstructionDiscriminant()
    REMOVE_MARKET_PRODUCT = InstructionDiscriminant()
    REMOVE_MARKET_PRODUCT_GROUP = InstructionDiscriminant()
    LOCK_COLLATERAL = InstructionDiscriminant()
    INITIALIZE_PRINT_TRADE = InstructionDiscriminant()
    SIGN_PRINT_TRADE = InstructionDiscriminant()
    INITIALIZE_TRADER_RISK_GROUP = InstructionDiscriminant()
    CLOSE_TRADER_RISK_GROUP = InstructionDiscriminant()
    NEW_ORDER = InstructionDiscriminant()
    NEW_ORDER_V2 = InstructionDiscriminant()
    MULTIPLACE = InstructionDiscriminant()
    MULTIPLACE_V2 = InstructionDiscriminant()
    CLEAR_OPEN_ORDERS = InstructionDiscriminant()
    REINITIALIZE_TRADER_POSITIONS = InstructionDiscriminant()
    SET_NUM_RISK_STATE_ACCOUNTS = InstructionDiscriminant()
    UPDATE_VARIANCE_CACHE = InstructionDiscriminant()
    CONSUME_ORDERBOOK_EVENTS = InstructionDiscriminant()
    CANCEL_ORDER = InstructionDiscriminant()
    CANCEL_ALL = InstructionDiscriminant()
    DEPOSIT_FUNDS = InstructionDiscriminant()
    WITHDRAW_FUNDS = InstructionDiscriminant()
    UPDATE_PRODUCT_FUNDING = InstructionDiscriminant()
    UPDATE_PRODUCT_MARK_PRICE_CONFIG = InstructionDiscriminant()
    TRANSFER_FULL_POSITION = InstructionDiscriminant()
    INITIALIZE_COMBO = InstructionDiscriminant()
    UPDATE_TRADER_FUNDING = InstructionDiscriminant()
    CLEAR_EXPIRED_ORDERBOOK = InstructionDiscriminant()
    POP_EVENTS = InstructionDiscriminant()
    SWEEP_FEES = InstructionDiscriminant()
    CHOOSE_SUCCESSOR = InstructionDiscriminant()
    CLAIM_AUTHORITY = InstructionDiscriminant()
    SETUP_CAPITAL_LIMITS = InstructionDiscriminant()
    UPDATE_CAPITAL_LIMITS = InstructionDiscriminant()
    UPDATE_TRADER_RISK_GROUP = InstructionDiscriminant()
    UPDATE_TRADER_RISK_GROUP_OWNER = InstructionDiscriminant()
    DISABLE_KILLSWITCH = InstructionDiscriminant()
    ENABLE_KILLSWITCH = InstructionDiscriminant()
    # LOCK-END





import asyncio
import argparse
import base64

from jsonrpcserver import Error
from dexteritysdk.dummy_oracle.state.oracle_price import OraclePrice
from dexteritysdk.dummy_oracle.api import OracleAPI
from dexteritysdk.program_ids import ORACLE_PROGRAM_ID
from podite import pod, U64, I64
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc.commitment import *
import solana.rpc.types as types
from solders.sysvar import SYSVAR_CLOCK_PUBKEY
import pandas as pd
import numpy as np
import datetime
import json
import websockets as ws
from dexteritysdk.utils import binance_parser


np.random.seed(100)

ap = argparse.ArgumentParser()
ap.add_argument("binance_ticker")
args = ap.parse_args()

bparser = binance_parser.BinanceParser(args.binance_ticker)


clock_key, _ = Pubkey.find_program_address(
    seeds = [
        b"clock",
    ],
    program_id=ORACLE_PROGRAM_ID,
)
oracle_key, _ = Pubkey.find_program_address(
        seeds=[
            b"oracle",
        ],
        program_id=ORACLE_PROGRAM_ID,
    )

print(oracle_key)
print(clock_key)

class Oracle:
    auth = Keypair(b"GOD " * 8)
    oracle_pubkey = oracle_key
    clock_pubkey = clock_key


@pod
class Clock:
    slot: U64
    epoch_start_time: I64
    epoch: U64
    leader_schedule_epoch: U64
    unix_timestamp: I64


# TODO: All of these should be command line args
URL = "http://127.0.0.1:8899"
WS_URI = "ws://127.0.0.1:8900"
ADJ = 100000


oracle_api = OracleAPI(rpc_url=URL)
oracle_api.create_client().request_airdrop(Oracle.auth.pubkey(), int(50 * 1e9))
print(f"Airdropped SOL to {str(Oracle.auth.pubkey())}")


def get_rpc_payload(key):
    return {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "accountSubscribe",
        "params": [
            key,
            {"encoding": "base64", "commitment": "confirmed"},
        ],
    }


clock_payload = get_rpc_payload(str(SYSVAR_CLOCK_PUBKEY))
oracle_payload = get_rpc_payload(str(Oracle.oracle_pubkey))
auth_payload = get_rpc_payload(str(Oracle.auth.pubkey()))


async def run():
    while True:
        try:
            clock = oracle_api._get_account_data_as(SYSVAR_CLOCK_PUBKEY, Clock)
            book = bparser.get_book()
            bbbo = book.iloc[0]
            bbest_bid = float(bbbo.bid)
            bbest_ask = float(bbbo.ask)
            current_price = int((bbest_ask + bbest_bid)/2)
            break
        except:
            pass

    clock = oracle_api._get_account_data_as(SYSVAR_CLOCK_PUBKEY, Clock)
    try:
        current_oracle = oracle_api.fetch_oracle_price(Oracle.oracle_pubkey)
        if current_oracle.tag._name_ == "UNINITIALIZED":
            raise Error()
        else:
            print(current_oracle)
            pass
    except:
        print("oracle: None; create a new oracle")
        (
            oracle_api.session(Oracle.auth)
            .initialize_oracle(Oracle.oracle_pubkey, Oracle.auth, current_price)
            .send(
                opts=types.TxOpts(
                    skip_preflight=True,
                    skip_confirmation=False,
                    preflight_commitment=Finalized,
                )
            )
        )
        current_oracle = oracle_api.fetch_oracle_price(Oracle.oracle_pubkey)

    try:
        current_clock = oracle_api.fetch_clock(Oracle.clock)
    except:
        print("clock: None; create a new clock")
        (
            oracle_api.session(Oracle.auth)
            .initialize_clock(Oracle.clock_pubkey, Oracle.auth)
            .send(
                opts=types.TxOpts(
                    skip_preflight=True,
                    skip_confirmation=False,
                    preflight_commitment=Finalized,
                )
            )
        )
        current_clock = oracle_api.fetch_clock(Oracle.clock_pubkey)
    while True:
        async with ws.connect(WS_URI) as w:
            await w.send(json.dumps(clock_payload))
            await w.send(json.dumps(oracle_payload))
            await w.send(json.dumps(auth_payload))
            resp = await (w.recv())
            clock_sub_id = json.loads(resp)["result"]
            resp = await (w.recv())
            oracle_sub_id = json.loads(resp)["result"]
            resp = await (w.recv())
            auth_sub_id = json.loads(resp)["result"]
            while True:
                try:
                    acct = await (w.recv())
                except:
                    break
                payload = json.loads(acct)
                if payload["params"]["subscription"] == clock_sub_id:
                    print(f"clock_sub_id: {clock_sub_id}")
                    data = base64.b64decode(
                        payload["params"]["result"]["value"]["data"][0]
                    )
                    clock = Clock.from_bytes(data)
                    while True:
                        try:
                            book = bparser.get_book()
                            bbbo = book.iloc[0]
                            bbest_bid = float(bbbo.bid)
                            bbest_ask = float(bbbo.ask)
                            price = int((bbest_ask + bbest_bid)/2)
                            break
                        except:
                            pass
                    current_oracle = oracle_api.fetch_oracle_price(Oracle.oracle_pubkey)
                    current_clock = oracle_api.fetch_clock(Oracle.clock_pubkey)
                    print(
                        f"{datetime.datetime.fromtimestamp(clock.unix_timestamp)}"
                        f" Oracle Price: {current_oracle.price}"
                        f" Oracle Clock slot {current_clock.slot}"
                        f" Update to binance price: {price}"
                        f" Oracle Slot: {current_oracle.slot}"
                        f" Clock Slot: {clock.slot}"
                    )
                    print(f"updating the price to: {price} and the clock.")
                    try:
                        (
                            oracle_api.session(Oracle.auth)
                            .update_price(Oracle.oracle_pubkey, Oracle.auth, price)
                            .update_clock(Oracle.clock_pubkey, Oracle.auth, clock.slot, clock.epoch_start_time, clock.epoch, clock.leader_schedule_epoch, clock.unix_timestamp)
                            .send()
                        )
                    except:
                        print("update price error")
                        break
                elif payload["params"]["subscription"] == oracle_sub_id:
                    print(f"oracle_sub_id: {oracle_sub_id}")
                    data = base64.b64decode(
                        payload["params"]["result"]["value"]["data"][0]
                    )
                    current_oracle = OraclePrice.from_bytes(data)
                elif payload["params"]["subscription"] == auth_sub_id:
                    print(f"auth_sub_id: {auth_sub_id}")
                    if payload["params"]["result"]["value"]["lamports"] <= 5000:
                        print('request airdrop')
                        try:
                            oracle_api.create_client().request_airdrop(
                                Oracle.auth.pubkey(), int(5 * 1e9)
                            )
                        except:
                            print('request failed')
                            break


asyncio.run(run())



from hashlib import sha256
import sys


def main():
    input = sys.argv[1]
    print(sighash(input).hex())
    print(sighash_int(input))


def sighash(ix_name: str) -> bytes:
    """Not technically sighash, since we don't include the arguments.
    (Because Rust doesn't allow function overloading.)
    Args:
        ix_name: The instruction name.
    Returns:
        The sighash bytes.
    """
    formatted_str = f"global:{ix_name}"
    return sha256(formatted_str.encode()).digest()[:8]


def sighash_int(ix_name: str) -> int:
    return int.from_bytes(sighash(ix_name), byteorder="little")


if __name__ == "__main__":
    main()



# LOCK-BEGIN[imports]: DON'T MODIFY
from .close_mark_prices import (
    CloseMarkPricesIx,
    close_mark_prices,
)
from .create_risk_state_account import (
    CreateRiskStateAccountIx,
    create_risk_state_account,
)
from .initialize_covariance_matrix import (
    InitializeCovarianceMatrixIx,
    initialize_covariance_matrix,
)
from .initialize_mark_prices import (
    InitializeMarkPricesIx,
    initialize_mark_prices,
)
from .instruction_tag import InstructionTag
from .remove_market_product_index_from_variance_cache import (
    RemoveMarketProductIndexFromVarianceCacheIx,
    remove_market_product_index_from_variance_cache,
)
from .resize_variance_cache import (
    ResizeVarianceCacheIx,
    resize_variance_cache,
)
from .update_mark_prices import (
    UpdateMarkPricesIx,
    update_mark_prices,
)
from .validate_account_health import (
    ValidateAccountHealthIx,
    validate_account_health,
)
from .validate_account_liquidation import (
    ValidateAccountLiquidationIx,
    validate_account_liquidation,
)

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.risk_engine.types.correlation_matrix import CorrelationMatrix
from dexteritysdk.codegen.risk_engine.types.covariance_metadata import CovarianceMetadata
from dexteritysdk.codegen.risk_engine.types.mark_prices_array import MarkPricesArray
from dexteritysdk.codegen.risk_engine.types.variance_cache import VarianceCache
from dexteritysdk.solmate.anchor import AccountDiscriminant
from podite import (
    Enum,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[accounts]: DON'T MODIFY
@pod
class Accounts(Enum[U64]):
    CORRELATION_MATRIX = AccountDiscriminant(field=CorrelationMatrix)
    COVARIANCE_METADATA = AccountDiscriminant(field=CovarianceMetadata)
    MARK_PRICES_ARRAY = AccountDiscriminant(field=MarkPricesArray)
    VARIANCE_CACHE = AccountDiscriminant(field=VarianceCache)
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import podite

from .sighash import sighash
from ..utils import snake_to_pascal


class AccountDiscriminant(podite.Variant):
    def assign_value(self, cls, prev_value):
        self.value = sighash("account", snake_to_pascal(self.name.lower()))


class InstructionDiscriminant(podite.Variant):
    def assign_value(self, cls, prev_value):
        self.value = sighash("global", self.name.lower())



from .base import *
from .event_queue import *
from .market_state import *
from .slab import *


def account_parser(data):
    tag = AccountTag.from_bytes(data[:1], byteorder="little")
    if tag == AccountTag.UNINITIALIZED:
        return None
    elif tag == AccountTag.MARKET:
        return MarketState.from_bytes(data)
    elif tag == AccountTag.EVENT_QUEUE:
        return EventQueue.from_bytes(data)
    elif tag == AccountTag.BIDS:
        return Slab.from_bytes(data)
    elif tag == AccountTag.ASKS:
        return Slab.from_bytes(data)
    raise ValueError()



# LOCK-BEGIN[imports]: DON'T MODIFY
from .cancel_all import (
    CancelAllIx,
    cancel_all,
)
from .cancel_order import (
    CancelOrderIx,
    cancel_order,
)
from .choose_successor import (
    ChooseSuccessorIx,
    choose_successor,
)
from .claim_authority import (
    ClaimAuthorityIx,
    claim_authority,
)
from .clear_expired_orderbook import (
    ClearExpiredOrderbookIx,
    clear_expired_orderbook,
)
from .close_trader_risk_group import (
    CloseTraderRiskGroupIx,
    close_trader_risk_group,
)
from .consume_orderbook_events import (
    ConsumeOrderbookEventsIx,
    consume_orderbook_events,
)
from .deactivate_market_product import (
    DeactivateMarketProductIx,
    deactivate_market_product,
)
from .deposit_funds import (
    DepositFundsIx,
    deposit_funds,
)
from .disable_killswitch import (
    DisableKillswitchIx,
    disable_killswitch,
)
from .enable_killswitch import (
    EnableKillswitchIx,
    enable_killswitch,
)
from .initialize_combo import (
    InitializeComboIx,
    initialize_combo,
)
from .initialize_market_product import (
    InitializeMarketProductIx,
    initialize_market_product,
)
from .initialize_market_product_group import (
    InitializeMarketProductGroupIx,
    initialize_market_product_group,
)
from .initialize_trader_risk_group import (
    InitializeTraderRiskGroupIx,
    initialize_trader_risk_group,
)
from .instruction_tag import InstructionTag
from .multiplace import (
    MultiplaceIx,
    multiplace,
)
from .multiplace_v2 import (
    MultiplaceV2Ix,
    multiplace_v2,
)
from .new_order import (
    NewOrderIx,
    new_order,
)
from .new_order_v2 import (
    NewOrderV2Ix,
    new_order_v2,
)
from .pop_events import (
    PopEventsIx,
    pop_events,
)
from .remove_market_product import (
    RemoveMarketProductIx,
    remove_market_product,
)
from .remove_market_product_group import (
    RemoveMarketProductGroupIx,
    remove_market_product_group,
)
from .set_num_risk_state_accounts import (
    SetNumRiskStateAccountsIx,
    set_num_risk_state_accounts,
)
from .setup_capital_limits import (
    SetupCapitalLimitsIx,
    setup_capital_limits,
)
from .sweep_fees import (
    SweepFeesIx,
    sweep_fees,
)
from .transfer_full_position import (
    TransferFullPositionIx,
    transfer_full_position,
)
from .update_capital_limits import (
    UpdateCapitalLimitsIx,
    update_capital_limits,
)
from .update_market_product_group import (
    UpdateMarketProductGroupIx,
    update_market_product_group,
)
from .update_product_funding import (
    UpdateProductFundingIx,
    update_product_funding,
)
from .update_product_mark_price_config import (
    UpdateProductMarkPriceConfigIx,
    update_product_mark_price_config,
)
from .update_trader_funding import (
    UpdateTraderFundingIx,
    update_trader_funding,
)
from .update_trader_risk_group import (
    UpdateTraderRiskGroupIx,
    update_trader_risk_group,
)
from .update_trader_risk_group_owner import (
    UpdateTraderRiskGroupOwnerIx,
    update_trader_risk_group_owner,
)
from .update_variance_cache import (
    UpdateVarianceCacheIx,
    update_variance_cache,
)
from .withdraw_funds import (
    WithdrawFundsIx,
    withdraw_funds,
)

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.account_tag import AccountTag
from dexteritysdk.codegen.dex.types.bitset import Bitset
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.product_array import ProductArray
from podite import (
    FixedLenArray,
    I16,
    U128,
    U16,
    U64,
    U8,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END

from collections.abc import Iterable
from dexteritysdk.codegen.dex import types


# LOCK-BEGIN[class(MarketProductGroup)]: DON'T MODIFY
@pod
class MarketProductGroup:
    tag: AccountTag
    name: FixedLenArray[U8, 16]
    authority: Pubkey
    successor: Pubkey
    vault_mint: Pubkey
    collected_fees: Fractional
    fee_collector: Pubkey
    decimals: U64
    risk_engine_program_id: Pubkey
    fee_model_program_id: Pubkey
    fee_model_configuration_acct: Pubkey
    risk_model_configuration_acct: Pubkey
    active_flags_products: Bitset
    ewma_windows: FixedLenArray[U64, 4]
    market_products: "ProductArray"
    vault_bump: U16
    risk_and_fee_bump: U16
    find_fees_discriminant_len: U16
    validate_account_discriminant_len: U16
    find_fees_discriminant: FixedLenArray[U8, 8]
    validate_account_health_discriminant: FixedLenArray[U8, 8]
    validate_account_liquidation_discriminant: FixedLenArray[U8, 8]
    create_risk_state_account_discriminant: FixedLenArray[U8, 8]
    max_maker_fee_bps: I16
    min_maker_fee_bps: I16
    max_taker_fee_bps: I16
    min_taker_fee_bps: I16
    fee_output_register: Pubkey
    risk_output_register: Pubkey
    sequence_number: U128
    staking_fee_collector: Pubkey
    is_killed: bool
    create_fee_state_account_discriminant: FixedLenArray[U8, 8]
    address_lookup_table: Pubkey
    close_risk_state_account_discriminant: FixedLenArray[U8, 8]
    close_fee_state_account_discriminant: FixedLenArray[U8, 8]
    is_admin_mode: bool
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)

    def active_products(self) -> Iterable[int, "types.Product"]:
        for idx, p in enumerate(self.market_products.array):
            if self.is_active(p):
                yield idx, p

    def is_active(self, product: "types.Product") -> bool:
        for p in self.market_products.array:
            if p.metadata().product_key != product.metadata().product_key:
                continue
            if p.is_outright():
                return p.field.outright.is_active()
            else:
                return all(
                    map(
                        lambda leg: self.market_products.array[leg.product_index].field.outright.is_active(),
                        p.field.combo.legs_array
                    )
                )

    def is_expired(self, product: "types.Product") -> bool:
        for p in self.market_products.array:
            if p.metadata().product_key != product.metadata().product_key:
                continue
            if p.is_outright():
                return p.field.outright.is_expired()
            else:
                return any(
                    map(
                        lambda leg: self.market_products.array[leg.product_index].field.outright.is_expired() or
                                    not self.market_products.array[leg.product_index].field.outright.is_active(),
                        p.field.combo.legs_array
                    )
                )


SENTINAL_KEY = Pubkey.from_string("11111111111111111111111111111111")



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.health_info import HealthInfo
from dexteritysdk.codegen.dex.types.liquidation_info import LiquidationInfo
from podite import (
    AutoTagType,
    Enum,
    Option,
    Variant,
    named_fields,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(HealthResult)]: DON'T MODIFY
@pod
class HealthResult(Enum[AutoTagType]):
    HEALTH = Variant(field=named_fields(health_info=HealthInfo))
    LIQUIDATION = Variant(field=named_fields(liquidation_info=LiquidationInfo))
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.risk_engine.types.fast_int import FastInt
from dexteritysdk.codegen.risk_engine.types.risk_account_tag import RiskAccountTag
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    FixedLenArray,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(VarianceCache)]: DON'T MODIFY
@pod
class VarianceCache:
    tag: RiskAccountTag
    update_slot: U64
    position_value: FastInt
    total_variance: FastInt
    open_order_variance: FastInt
    product_indexes: FixedLenArray[Usize, 32]
    positions: FixedLenArray[FastInt, 32]
    sigma_p: FixedLenArray[FastInt, 32]
    total_liquidity_buffer: FastInt
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from dexteritysdk.program_ids import ORACLE_PROGRAM_ID
from dexteritysdk.dummy_oracle.state import OraclePrice, Clock

from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.transaction import Transaction
from solders.system_program import create_account, CreateAccountParams

import dexteritysdk.dummy_oracle.instructions as ixs

from dexteritysdk.utils.solana import (
    actionify,
    Context,
)


def _calc_rent(space, client=None):
    if client is None:
        client = Context.get_global_client()
    return client.get_minimum_balance_for_rent_exemption(space)["result"]


def extract_acct_addr(resp, idx=0):
    addr = resp.instructions[0]["accounts"][idx]
    exists = False
    if resp.error:
        error_ix, error_info = resp.error["InstructionError"]
        if error_ix == 0 and error_info["Custom"] == 0:
            exists = True
    else:
        exists = True

    if exists:
        return addr, resp
    else:
        return None, resp


@actionify
def initialize_oracle(
    authority: Pubkey,
    oracle: Pubkey,
    price: int,
    decimals: int = 0,
):
    space = OraclePrice.calc_size()
    rent = _calc_rent(space)
    return Transaction().add(
        create_account(
            CreateAccountParams(
                from_pubkey=authority,
                new_account_pubkey=oracle,
                lamports=rent,
                space=space,
                program_id=ORACLE_PROGRAM_ID,
            )
        ),
        ixs.initialize_oracle_ix(
            oracle_price=oracle,
            update_authority=authority,
            price=price,
            decimals=decimals,
        ),
    )


@actionify
def initialize_clock(
    authority: Pubkey,
    clock: Pubkey,
):
    space = Clock.calc_size()
    rent = _calc_rent(space)
    return Transaction().add(
        create_account(
            CreateAccountParams(
                from_pubkey=authority,
                new_account_pubkey=clock,
                lamports=rent,
                space=space,
                program_id=ORACLE_PROGRAM_ID,
            )
        ),
        ixs.initialize_clock_ix(
            clock=clock,
            update_authority=authority,
        ),
    )


@actionify
def update_clock(
    clock: Pubkey,
    slot: int = 0,
    epoch_start_timestamp: int = 0,
    epoch: int = 0,
    leader_schedule_epoch: int = 0,
    unix_timestamp: int = 0,
):
    return Transaction().add(
        ixs.update_clock_ix(
            clock=clock,
            slot=slot,
            epoch_start_timestamp=epoch_start_timestamp,
            epoch=epoch,
            leader_schedule_epoch=leader_schedule_epoch,
            unix_timestamp=unix_timestamp,
        )
    )


@actionify
def update_price(
    oracle: Pubkey,
    authority: Keypair,
    price: int,
    decimals: int = 0,
):
    return Transaction().add(
        ixs.update_price_ix(
            oracle_price=oracle,
            update_authority=authority,
            price=price,
            decimals=decimals,
        )
    )



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.capital_limits import CapitalLimits
from dexteritysdk.codegen.dex.types.capital_limits_params import CapitalLimitsParams
from dexteritysdk.codegen.dex.types.market_product_group import MarketProductGroup
from dexteritysdk.codegen.dex.types.product_array import ProductArray
from dexteritysdk.codegen.dex.types.risk_output_register import RiskOutputRegister
from dexteritysdk.codegen.dex.types.trader_risk_group import TraderRiskGroup
from dexteritysdk.solmate.anchor import AccountDiscriminant
from podite import (
    Enum,
    U64,
    pod,
)

# LOCK-END

from podite import AutoTagTypeValueManager

# LOCK-BEGIN[accounts]: DON'T MODIFY
@pod
class Accounts(Enum[U64]):
    CAPITAL_LIMITS = AccountDiscriminant(field=CapitalLimits)
    MARKET_PRODUCT_GROUP = AccountDiscriminant(field=MarketProductGroup)
    PRODUCT_ARRAY = AccountDiscriminant(field=ProductArray)
    RISK_OUTPUT_REGISTER = AccountDiscriminant(field=RiskOutputRegister)
    TRADER_RISK_GROUP = AccountDiscriminant(field=TraderRiskGroup)
    CAPITAL_LIMITS_PARAMS = AccountDiscriminant(field=CapitalLimitsParams)
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", format="FORMAT_ZERO_COPY", **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, format="FORMAT_ZERO_COPY", **kwargs):
        # accounts don't have the same size variants, so must manually use zero-copy and
        # set auto tag type.
        # TODO make this api cleaner
        with AutoTagTypeValueManager(U64):
            return super()._inner_from_bytes_partial(buffer, format="FORMAT_ZERO_COPY", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.instruments.types.account_tag import AccountTag
from dexteritysdk.codegen.instruments.types.expiration_status import ExpirationStatus
from dexteritysdk.codegen.instruments.types.instrument_type import InstrumentType
from dexteritysdk.codegen.instruments.types.oracle_type import OracleType
from dexteritysdk.solmate.dtypes import UnixTimestamp
from podite import (
    U64,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(DerivativeMetadata)]: DON'T MODIFY
@pod
class DerivativeMetadata:
    tag: AccountTag
    expiration_status: ExpirationStatus
    oracle_type: OracleType
    instrument_type: InstrumentType
    bump: U64
    strike: Fractional
    initialization_time: UnixTimestamp
    full_funding_period: UnixTimestamp
    minimum_funding_period: UnixTimestamp
    price_oracle: Pubkey
    market_product_group: Pubkey
    close_authority: Pubkey
    clock: Pubkey
    last_funding_time: UnixTimestamp
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from io import BytesIO
from podite import BYTES_CATALOG
from solders.pubkey import Pubkey
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from dexteritysdk.solmate.utils import to_account_meta
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(validate_account_liquidation)]: DON'T MODIFY
@dataclass
class ValidateAccountLiquidationIx:
    program_id: Pubkey

    # account metas
    market_product_group: AccountMeta
    trader_risk_group: AccountMeta
    out_register_risk_info: AccountMeta
    risk_state: AccountMeta
    risk_model_configuration: AccountMeta
    risk_signer: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    def to_instruction(self):
        keys = []
        keys.append(self.market_product_group)
        keys.append(self.trader_risk_group)
        keys.append(self.out_register_risk_info)
        keys.append(self.risk_state)
        keys.append(self.risk_model_configuration)
        keys.append(self.risk_signer)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.VALIDATE_ACCOUNT_LIQUIDATION))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(validate_account_liquidation)]: DON'T MODIFY
def validate_account_liquidation(
    market_product_group: Union[str, Pubkey, AccountMeta],
    trader_risk_group: Union[str, Pubkey, AccountMeta],
    out_register_risk_info: Union[str, Pubkey, AccountMeta],
    risk_state: Union[str, Pubkey, AccountMeta],
    risk_model_configuration: Union[str, Pubkey, AccountMeta],
    risk_signer: Union[str, Pubkey, AccountMeta],
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("X5CpjB92BKHb7EGHsBRTBapbVpz2iGNGnM1ekNyDhMB")

    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(trader_risk_group, (str, Pubkey)):
        trader_risk_group = to_account_meta(
            trader_risk_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(out_register_risk_info, (str, Pubkey)):
        out_register_risk_info = to_account_meta(
            out_register_risk_info,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_state, (str, Pubkey)):
        risk_state = to_account_meta(
            risk_state,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_model_configuration, (str, Pubkey)):
        risk_model_configuration = to_account_meta(
            risk_model_configuration,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_signer, (str, Pubkey)):
        risk_signer = to_account_meta(
            risk_signer,
            is_signer=True,
            is_writable=False,
        )

    return ValidateAccountLiquidationIx(
        program_id=program_id,
        market_product_group=market_product_group,
        trader_risk_group=trader_risk_group,
        out_register_risk_info=out_register_risk_info,
        risk_state=risk_state,
        risk_model_configuration=risk_model_configuration,
        risk_signer=risk_signer,
        remaining_accounts=remaining_accounts,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.multiplace_param import MultiplaceParam
from dexteritysdk.utils.aob.state.base import SelfTradeBehavior
from podite import (
    U32,
    U64,
    U8,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(MultiplaceV2Params)]: DON'T MODIFY
@pod
class MultiplaceV2Params:
    match_limit: U64
    referrer_fee_bps: Fractional
    self_trade_behavior: SelfTradeBehavior
    expiration_slot: U32
    num_new_orders: U8
    new_order0: "MultiplaceParam"
    new_order1: "MultiplaceParam"
    new_order2: "MultiplaceParam"
    new_order3: "MultiplaceParam"
    new_order4: "MultiplaceParam"
    new_order5: "MultiplaceParam"
    new_order6: "MultiplaceParam"
    new_order7: "MultiplaceParam"
    new_order8: "MultiplaceParam"
    new_order9: "MultiplaceParam"
    new_order10: "MultiplaceParam"
    new_order11: "MultiplaceParam"
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from abc import ABC
from dataclasses import dataclass

from solders.pubkey import Pubkey

from dexteritysdk.pyserum.layouts.event_queue import Side, CompletedReason


@dataclass
class EventQueueHeader:
    tag: int
    head: int
    count: int
    event_size: int
    seq_num: int

    @staticmethod
    def from_construct(c_obj):
        return EventQueueHeader(
            c_obj.tag,
            c_obj.head,
            c_obj.count,
            c_obj.event_size,
            c_obj.seq_num
        )


@dataclass
class Register:
    posted_order_id: int
    total_base_quantity: int
    total_quote_quantity: int
    total_base_quantity_posted: int

    @staticmethod
    def from_construct(c_obj):
        if c_obj.register_present:
            return Register(
                c_obj.posted_order_id if c_obj.posted_order_id_present else None,
                c_obj.total_base_qty,
                c_obj.total_quote_qty,
                c_obj.total_base_qty_posted
            )
        return None


@dataclass
class Callback:
    account: Pubkey
    open_orders_idx: int
    order_nonce: int
    client_order_id: int

    @staticmethod
    def from_construct(c_obj):
        return Callback(
            Pubkey.from_bytes(c_obj.account),
            c_obj.open_orders_idx,
            c_obj.order_nonce,
            c_obj.client_order_id,
        )


class Event(ABC):
    pass


@dataclass
class FillEvent(Event):
    taker_side: Side
    maker_order_id: int
    quote_size: int
    base_size: int
    maker_callback: Callback
    taker_callback: Callback

    @staticmethod
    def from_construct(c_obj):
        return FillEvent(
            c_obj.taker_side,
            c_obj.maker_order_id,
            c_obj.quote_size,
            c_obj.base_size,
            Callback.from_construct(c_obj.maker_callback),
            Callback.from_construct(c_obj.taker_callback),
        )


@dataclass
class OutEvent(Event):
    side: Side
    order_id: int
    base_size: int
    delete: int
    callback: Callback
    reason: CompletedReason

    @staticmethod
    def from_construct(c_obj):
        return OutEvent(
            c_obj.side,
            c_obj.order_id,
            c_obj.base_size,
            c_obj.delete,
            Callback.from_construct(c_obj.callback),
            c_obj.reason,
        )



import os
import json
import re
import subprocess
from functools import partial
from typing import Callable, Dict, Union, Set

import solders
from solders.pubkey import Pubkey
from dexteritysdk.solmate.anchor import Idl
from dexteritysdk.solmate.anchor.codegen import usize_type, unix_timestamp_type, program_error_type, CodeGen
from dexteritysdk.solmate.anchor.editor import CodeEditor
from dexteritysdk.solmate.utils import pascal_to_snake
from spl.token.constants import TOKEN_PROGRAM_ID

from dexteritysdk.scripts.extract_program_ids import get_root


def main():
    root = get_root()
    cli(idl_dir=f'{root}/target/idl', out_dir=f'{root}/client', pids_dir=f'{root}/target/deploy',
        parent_module='dexteritysdk.codegen', skip_types={"TwoIterators", "DomainOrProgramError"})


if __name__ == '__main__':
    main()


def side_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.utils.aob.state.base", "Side")
    return "Side"


def self_trade_behavior_type(editor: CodeEditor):
    editor.add_from_import("dexteritysdk.utils.aob.state.base", "SelfTradeBehavior")
    return "SelfTradeBehavior"


def cli(
        idl_dir: str,
        out_dir: str,
        pids_dir: Union[str, Dict[str, Pubkey]],
        parent_module: str,
        skip_types: Set[str],
):
    protocol_to_idl_and_types = {}
    protocol_to_pid = get_protocols(idl_dir, pids_dir)
    for protocol in protocol_to_pid.keys():
        idl = Idl.from_json_file(f"{idl_dir}/{protocol}.json")

        idl.types = list(filter(lambda x: x.name not in skip_types, idl.types))
        idl.accounts = list(filter(lambda x: x.name not in skip_types, idl.accounts))
        protocol_to_idl_and_types[protocol] = (
            idl,
            defined_types_to_imports(f"{parent_module}.{protocol}", idl),
        )

        if protocol.upper() in os.environ:
            env_pid = os.environ[protocol.upper()]
            print(f"Using program_id from ENV for {protocol}: {env_pid}")
            protocol_to_pid[protocol] = env_pid
        else:
            print(f"Using program_id from target/deploy/{protocol}-keypair.json for {protocol}: "
                  f"{protocol_to_pid[protocol]}")

        if protocol_to_pid[protocol] is None:
            raise Exception("Found IDL but no program id set")

    external_types = {
        "usize": usize_type,
        "UnixTimestamp": unix_timestamp_type,
        "ProgramError": program_error_type,
        "SelfTradeBehavior": self_trade_behavior_type,
        "Side": side_type,
    }
    # allow each IDL to reference types defined in other IDLs
    for (_, exported_types) in protocol_to_idl_and_types.values():
        external_types.update(exported_types)

    for protocol, (idl, _) in protocol_to_idl_and_types.items():
        print(f"Generating code for {protocol}")
        codegen = CodeGen(
            idl=idl,
            addresses={},
            root_module=f"{parent_module}.{protocol}",
            source_path=out_dir,
            external_types=external_types,
            default_accounts={
                "systemProgram": solders.system_program.ID,
                "token_program": Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"),
                "sysvar_rent": solders.sysvar.RENT,
                "program_id": protocol_to_pid[protocol],
            },
            accnt_tag_values="anchor",
            instr_tag_values="anchor",
            skip_types=skip_types,
        )
        codegen.generate_code(check_missing_types=not True)
        codegen.save_modules()


def defined_types_to_imports(
        root_module: str, idl: Idl
) -> Dict[str, Callable[[CodeEditor], str]]:
    def add_import(name: str, editor: CodeEditor) -> str:
        editor.add_from_import(f"{root_module}.types.{pascal_to_snake(name)}", name)
        return name

    type_definitions = idl.types + idl.accounts
    return dict(((ty.name, partial(add_import, ty.name)) for ty in type_definitions))


def get_protocols(idl_dir: str, pids: Union[str, Dict[str, str]]) -> Dict[str, str]:
    protocols = set()
    for filename in os.listdir(idl_dir):
        match = re.search(r"([a-z_\-]+).json", filename)
        if match is None:
            continue
        protocol = match.groups()[0]
        protocols.add(protocol)

    # if a path was passed in, load pids from the directory
    if isinstance(pids, str):
        pids = dir_to_pids(pids)

    intersection = {}
    for protocol, pid in pids.items():
        if protocol in protocols:
            intersection[protocol] = pid
    for protocol in protocols:
        if protocol not in pids:
            print("WARNING: found idl file with no matching program id: ", protocol)
            intersection[protocol] = None

    return intersection


def dir_to_pids(dir: str) -> Dict[str, str]:
    program_to_id = {}
    root = run("git rev-parse --show-toplevel")
    for filename in os.listdir(dir):
        match = re.search(r"([a-z_]+)-keypair.json", filename)
        if match is None:
            continue
        program = match.groups()[0]
        program_to_id[program] = run(f"solana-keygen pubkey {dir}/{filename}")
    return program_to_id


def run(cmd, debug=False):
    if debug:
        print(cmd)
    res = subprocess.check_output(cmd, shell=True).strip().decode("utf-8")
    if debug:
        print(res)
    return res



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.utils.aob.state.base import Side
from podite import pod
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(TraderFeeParams)]: DON'T MODIFY
@pod
class TraderFeeParams:
    side: Side
    is_aggressor: bool
    matched_quote_qty: "Fractional"
    matched_base_qty: "Fractional"
    product: Pubkey
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from __future__ import annotations
from typing import NamedTuple
from dexteritysdk.pyserum.enums import Side


class OrderInfo(NamedTuple):
    price: float
    """"""
    size: float
    """"""
    trader: bytes
    """"""
    expiration_slot: int
    """"""


class Order(NamedTuple):
    order_id: int
    """"""
    client_id: int
    """"""
    info: OrderInfo
    """"""
    side: Side
    """"""



"""Serum specific enums."""

from enum import IntEnum


class Side(IntEnum):
    """Side of the orderbook to trade."""

    BUY = 0
    """"""
    SELL = 1
    """"""


class OrderType(IntEnum):
    """Type of order."""

    LIMIT = 0
    """"""
    IOC = 1
    """"""
    POST_ONLY = 2
    """"""


class SelfTradeBehavior(IntEnum):
    DECREMENT_TAKE = 0
    CANCEL_PROVIDE = 1
    ABORT_TRANSACTION = 2



import datetime
import sched
from threading import Lock, Thread
from binance import ThreadedWebsocketManager, BinanceSocketManager
import pandas as pd

pd.options.mode.chained_assignment = None


class BinanceParser:
    def __init__(self, symbol):
        # self.logger = logger
        self.exchange = "BINANCE"
        self.symbol = symbol
        self.socket_depth = BinanceSocketManager.WEBSOCKET_DEPTH_10
        self.depth = 10  # int(cfg.get("depth"), 10)
        self.book_lock = Lock()
        self.trade_lock = Lock()
        self.active_book = []
        self.active_trades = []
        self.last_book_update = datetime.datetime.utcnow()
        self.last_trade_update = datetime.datetime.utcnow()
        self.scheduler = sched.scheduler()
        self.heartbeat_thread = Thread(target=self.scheduler.run)
        self.scheduler.enter(10, 0, self._heartbeat)
        self.heartbeat_thread.start()
        self.binance_client = None
        self._start()

    def has_product(self):
        return self.symbol is not None

    def reconnect(self):
        if self.binance_client is not None:
            try:
                self.binance_client.stop()
            except:
                print("Failed to stop ThreadedWebSocketManager")
        self._start()

    def _heartbeat(self):
        self.scheduler.enter(10, 0, self._heartbeat)
        now = datetime.datetime.utcnow()
        if (now - self.last_book_update).seconds > 10:
            print(
                f"Likely stale Binance Order Book connection for {self.symbol}, reconnecting"
            )
            self.reconnect()
        elif (now - self.last_trade_update).seconds > 10:
            print(
                f"Likely stale Binance Trade connection for {self.symbol}, reconnecting"
            )
            self.reconnect()
        else:
            if False:
                print(f"Connection seems fine")

    def _start(self):
        if self.symbol is None:
            return
        self.binance_client = ThreadedWebsocketManager()
        self.binance_client.start()
        self.depth_stream = self.binance_client.start_depth_socket(
            callback=self._update_book, symbol=self.symbol, depth=self.depth
        )
        self.trade_stream = self.binance_client.start_trade_socket(
            callback=self._update_trades,
            symbol=self.symbol,
        )

    def _update_book(self, msg):
        self.last_book_update = datetime.datetime.utcnow()
        with self.book_lock:
            msg["timestamp"] = self.last_book_update
            self.active_book.append(msg)

    def _update_trades(self, msg):
        self.last_trade_update = datetime.datetime.utcnow()
        with self.trade_lock:
            self.active_trades.append(msg)

    def get_book(self):
        with self.book_lock:
            books = self.active_book
            self.active_book = []
        dfs = []
        for book in books:
            bids = pd.DataFrame(book["bids"], columns=["bid", "nbid"])
            asks = pd.DataFrame(book["asks"], columns=["ask", "nask"])
            df = (
                bids.join(asks)
                .head(self.depth)
                .reset_index()
                .rename(columns={"index": "level"})
            )
            df["timestamp"] = book["timestamp"]
            dfs.append(df)
        if not dfs:
            return pd.DataFrame()
        df = pd.concat(dfs)
        return df

    def get_trades(self, start, end):
        with self.trade_lock:
            trade_list = self.active_trades
            self.active_trades = []
        trade_data = []
        for trade in trade_list:
            trade_data.append(
                {
                    "order_id": trade["t"],
                    "timestamp": trade["T"],
                    "price": trade["p"],
                    "volume": trade["q"],
                    "side": '"sell"' if trade["m"] else '"buy"',
                }
            )
        trades = pd.DataFrame(trade_data)
        if not trades.empty:
            trades.timestamp = pd.to_datetime(
                trades.timestamp, unit="ms"
            ).dt.tz_localize("UTC")
            trades["timestamp"] = trades["timestamp"].dt.tz_localize(None)
            trades["query_start_time"] = int(start.timestamp() * 1e9)
            trades["query_end_time"] = int(end.timestamp() * 1e9)
        return trades

    def get_pool(self):
        return pd.DataFrame()



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.solmate.dtypes import Usize
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.pubkey import Pubkey
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(remove_market_product_index_from_variance_cache)]: DON'T MODIFY
@dataclass
class RemoveMarketProductIndexFromVarianceCacheIx:
    program_id: Pubkey

    # account metas
    authority: AccountMeta
    market_product_group: AccountMeta
    trader_risk_group: AccountMeta
    covariance_metadata: AccountMeta
    variance_cache: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    market_product_index: Usize

    def to_instruction(self):
        keys = []
        keys.append(self.authority)
        keys.append(self.market_product_group)
        keys.append(self.trader_risk_group)
        keys.append(self.covariance_metadata)
        keys.append(self.variance_cache)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.REMOVE_MARKET_PRODUCT_INDEX_FROM_VARIANCE_CACHE))
        buffer.write(BYTES_CATALOG.pack(Usize, self.market_product_index))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(remove_market_product_index_from_variance_cache)]: DON'T MODIFY
def remove_market_product_index_from_variance_cache(
    authority: Union[str, Pubkey, AccountMeta],
    market_product_group: Union[str, Pubkey, AccountMeta],
    trader_risk_group: Union[str, Pubkey, AccountMeta],
    covariance_metadata: Union[str, Pubkey, AccountMeta],
    variance_cache: Union[str, Pubkey, AccountMeta],
    market_product_index: Usize,
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("92wdgEqyiDKrcbFHoBTg8HxMj932xweRCKaciGSW3uMr")

    if isinstance(authority, (str, Pubkey)):
        authority = to_account_meta(
            authority,
            is_signer=True,
            is_writable=False,
        )
    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(trader_risk_group, (str, Pubkey)):
        trader_risk_group = to_account_meta(
            trader_risk_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(covariance_metadata, (str, Pubkey)):
        covariance_metadata = to_account_meta(
            covariance_metadata,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(variance_cache, (str, Pubkey)):
        variance_cache = to_account_meta(
            variance_cache,
            is_signer=False,
            is_writable=True,
        )

    return RemoveMarketProductIndexFromVarianceCacheIx(
        program_id=program_id,
        authority=authority,
        market_product_group=market_product_group,
        trader_risk_group=trader_risk_group,
        covariance_metadata=covariance_metadata,
        variance_cache=variance_cache,
        remaining_accounts=remaining_accounts,
        market_product_index=market_product_index,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from podite import (
    Vec,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(Health)]: DON'T MODIFY
@pod
class Health:
    margin_req: Fractional
    portfolio_value: Fractional
    total_abs_dollar_position: Fractional
    abs_dollar_position: Vec[Fractional]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from solders.pubkey import Pubkey
from solders.instruction import AccountMeta, Instruction
from dexteritysdk.program_ids import SYSTEM_PROGRAM_ID as SYS_PROGRAM_ID
from podite import I64, U64, pod

from .common import InstructionCode
from dexteritysdk.program_ids import ORACLE_PROGRAM_ID


@pod
class Params:
    instr: InstructionCode


def initialize_clock_ix(clock: Pubkey, update_authority: Pubkey):
    keys = [
        AccountMeta(pubkey=clock, is_signer=False, is_writable=True),
        AccountMeta(pubkey=update_authority, is_signer=True, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    params = Params(instr=InstructionCode.INITIALIZE_CLOCK)
    return Instruction(
        accounts=keys, program_id=ORACLE_PROGRAM_ID, data=params.to_bytes()
    )



import argparse
import json

from base58 import b58encode as b58e
from base58 import b58decode as b58d
from solders.keypair import Keypair

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("f")
    args = ap.parse_args()

    try:
        # try to open as file
        with open(args.f, "r") as f:
            kp = json.load(f)
        print(b58e(bytes(kp)).decode('ascii'))
        print(Keypair(kp[:32]).pubkey())
    except:
        try: 
            # try to interpret as b58
            kp = [byte for byte in b58d(args.f)]
            print(kp)
            print(kp[32:])
            print(b58e(bytes(kp[32:])).decode('ascii'))
        except:
            # try to interpret as list of bytes
            kp = [int(byte) for byte in args.f[1:-1].split(',')]
            print(kp)
            print(kp[32:])
            print(b58e(bytes(kp[32:])).decode('ascii'))
            print(b58e(bytes(kp)).decode('ascii'))

if __name__ == '__main__':
    main()



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    FixedLenArray,
    U128,
    U64,
    U8,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(CancelOrderParams)]: DON'T MODIFY
@pod
class CancelOrderParams:
    order_id: U128
    no_err: FixedLenArray[U8, 1]
    client_order_id: U64
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.order_type import OrderType
from dexteritysdk.utils.aob.state.base import (
    SelfTradeBehavior,
    Side,
)
from podite import (
    U32,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(NewOrderV2Params)]: DON'T MODIFY
@pod
class NewOrderV2Params:
    side: Side
    max_base_qty: Fractional
    order_type: "OrderType"
    self_trade_behavior: SelfTradeBehavior
    match_limit: U64
    limit_price: Fractional
    referrer_fee_bps: Fractional
    client_order_id: U64
    expiration_slot: U32
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import base58
import base64
import sys

# Function to search for bytes
def find_bytes(haystack, needle):
    return haystack.find(needle)

# Base58-decoded bytes
base58_encoded_str = sys.argv[2]
decoded_base58 = base58.b58decode(base58_encoded_str)

# Read and Base64-decode the file
with open(sys.argv[1], 'r') as file:
    base64_encoded_data = file.read()
    decoded_base64 = base64.b64decode(base64_encoded_data)

# Search for Base58-decoded bytes in Base64-decoded file
position = find_bytes(decoded_base64, decoded_base58)

print(position)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .correlation_matrix import CorrelationMatrix
from .covariance_metadata import CovarianceMetadata
from .fast_int import FastInt
from .initialize_mark_prices_params import InitializeMarkPricesParams
from .mark_price import MarkPrice
from .mark_prices_array import MarkPricesArray
from .risk_account_tag import RiskAccountTag
from .risk_error import RiskError
from .update_mark_prices_params import UpdateMarkPricesParams
from .variance_cache import VarianceCache

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from podite import pod

# LOCK-END


# LOCK-BEGIN[class(UpdateProductMarkPriceConfigParams)]: DON'T MODIFY
@pod
class UpdateProductMarkPriceConfigParams:
    mark_price_qualifying_cum_value: Fractional
    mark_price_max_qualifying_width: Fractional
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
import dexteritysdk.codegen.alpha_risk_engine.accounts as accounts
import dexteritysdk.codegen.alpha_risk_engine.instructions as instructions
import dexteritysdk.codegen.alpha_risk_engine.types as types

# LOCK-END





# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    AutoTagType,
    Enum,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(RiskError)]: DON'T MODIFY
@pod
class RiskError(Enum[AutoTagType]):
    INVALID_ACCOUNT_TAG = None
    ACCOUNT_ALREADY_INITIALIZED = None
    INVALID_RISK_SIGNER = None
    INVALID_ACCOUNT_OWNER = None
    INVALID_ACCOUNT_ADDRESS = None
    INVALID_COVARIANCE_AUTHORITY = None
    INVALID_COVARIANCE_MATRIX_ACCESS = None
    MISSING_COVARIANCE_ENTRY = None
    INVALID_SQRT_INPUT = None
    INVALID_COVARIANCE_INPUT = None
    MISSING_B_B_O_FOR_MARK_PRICE = None
    NUMERICAL_OVERFLOW = None
    UNEXPECTED_PRODUCT_TYPE = None
    UNEXPECTED_RESULT = None
    MISMATCHED_RISK_STATE_ACCOUNT = None
    FAILED_TO_FIND_CACHE_INDEX_FOR_LEG = None
    COMBO_SIZE_GREATER_THAN_COLLECTION_LEN = None
    INVALID_MARK_PRICE_ACCOUNTS_LEN = None
    MISMATCHED_ORACLE_PRICE_ACCOUNT = None
    MISSING_MARK_PRICE = None
    INCORRECT_MARK_PRICES_BUMP = None
    MARK_PRICES_ARRAY_IS_FULL = None
    MARK_PRICES_OUT_OF_DATE = None
    FAILED_TO_FIND_MARKET_PRODUCT_INDEX_IN_VARIANCE_CACHE = None
    BOOK_SPREAD_TOO_WIDE_FOR_MARK_PRICE = None
    # LOCK-END

    @classmethod
    def _to_bytes_partial(cls, buffer, obj, **kwargs):
        # to modify packing, change this method
        return super()._to_bytes_partial(buffer, obj, **kwargs)

    @classmethod
    def _from_bytes_partial(cls, buffer, **kwargs):
        # to modify unpacking, change this method
        return super()._from_bytes_partial(buffer, **kwargs)

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.risk_engine.types.fast_int import FastInt
from podite import (
    FixedLenArray,
    U64,
    U8,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(MarkPrice)]: DON'T MODIFY
@pod
class MarkPrice:
    product_key: Pubkey
    mark_price: FastInt
    prev_oracle_minus_book_ewma: FastInt
    oracle_minus_book_ewma: FastInt
    update_slot: U64
    is_qualifying_bid_price_some: U8
    _padding0: FixedLenArray[U8, 7]
    qualifying_bid_price: Fractional
    is_qualifying_ask_price_some: U8
    _padding1: FixedLenArray[U8, 7]
    qualifying_ask_price: Fractional
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import json
from random import random
import sys
import time

print('importing dexterity (it takes a while)...', end='', flush=True)
from dexteritysdk.dex.sdk_context import SDKContext, SDKTrader
from dexteritysdk.utils.aob import Side
print('success')
from solana.rpc.api import Client
from solders.keypair import Keypair
from solders.pubkey import Pubkey


CRANK_SLEEP_SECONDS = 5
BTC_PRICE = 69000
ETH_PRICE = 4000
SOL_PRICE = 150


def load_keypair(fp):
    with open(fp, "r") as f:
        kp = json.load(f)
    return Keypair.from_bytes(kp)

async def main():
    network = "http://localhost:8899/"
    client = Client(network)

    print('loading keypairs...', end='', flush=True)
    keypair = load_keypair('../deploy_key.json')
    mpg_keypair = load_keypair('../target/deploy/mpg-keypair.json')
    print('success')

    print('connecting...', end='', flush=True)
    ctx = SDKContext.connect(
	client=client,
	market_product_group_key=mpg_keypair.pubkey(),
	payer=keypair,
	raise_on_error=True
    )
    print('success')

    print('creating maker trg...', end='', flush=True)
    maker = ctx.register_trader(keypair)
    print('success (maker trg pubkey: {})'.format(maker.account))
    print('subscribing...', end='', flush=True)
    asyncio.create_task(maker.subscribe(ctx, 'ws://localhost:8900',
                                        fills_callback=lambda fill, event_type: print(f"saw fill event {event_type} {fill}")))
    await asyncio.sleep(1)
    print('success')

    print('creating taker trg...', end='', flush=True)
    taker = ctx.register_trader(keypair)
    asyncio.create_task(taker.subscribe(ctx, 'ws://localhost:8900'))
    print('success (taker trg pubkey: {})'.format(taker.account))

    INITIAL_DEPOSIT = 100000
    print('maker depositing 100k USDC...', end='', flush=True)
    maker.deposit(ctx, INITIAL_DEPOSIT)
    print('success')
    print('taker depositing 100k USDC...', end='', flush=True)
    taker.deposit(ctx, INITIAL_DEPOSIT)
    print('success')

    print('checking balance...', end='', flush=True)
    maker_trg, _ = maker.get_trader_risk_group()
    if maker_trg.cash_balance.value != INITIAL_DEPOSIT:
        raise RuntimeError(f"expected maker_trg.cash_balance == {INITIAL_DEPOSIT}, got {maker_trg.cash_balance}")
    taker_trg, _ = taker.get_trader_risk_group()
    if taker_trg.cash_balance.value != INITIAL_DEPOSIT:
        raise RuntimeError(f"expected taker_trg.cash_balance == {INITIAL_DEPOSIT}, got {taker_trg.cash_balance}")
    print('success')

    size, price = 0.1, BTC_PRICE

    product = ctx.products[0]

    client_order_id = int(random() * 1e6)
    print(f"placing two ASK orders {size} @ {price},{price+1} with client_order_id={client_order_id},{client_order_id+1} on {product.name}...", end='', flush=True)
    maker.place_order(ctx, product, Side.ASK, size, price, client_order_id=client_order_id)
    maker.place_order(ctx, product, Side.ASK, size, price+1, client_order_id=client_order_id+1)
    print('success')

    taker_size = size*2
    print(f"taking both orders...", end='', flush=True)
    taker.place_order(ctx, product, Side.BID, taker_size, price+1)
    print('success')

    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('checking open orders...', end='', flush=True)
    open_orders = list(maker.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero maker open orders but saw {len(open_orders)} open orders")
    open_orders = list(taker.open_orders(ctx))
    if len(open_orders) != 0:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected zero taker open orders but saw {len(open_orders)} open orders")
    print('success')

    client_order_id = int(random() * 1e6)
    print(f"placing ASK order {size} @ {price} with client_order_id={client_order_id} on {product.name}...", end='', flush=True)
    maker.place_order(ctx, product, Side.ASK, size, price, client_order_id=client_order_id)
    print('success')

    taker_size = size*0.25
    print(f"taking {size} @ {price}...", end='', flush=True)
    taker.place_order(ctx, product, Side.BID, taker_size, price)
    print('success')
    
    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')
    print('checking open orders...', end='', flush=True)
    open_orders = list(maker.open_orders(ctx))
    if len(open_orders) != 1:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected 1 open order but saw {len(open_orders)} open orders")
    if open_orders[0].client_order_id != client_order_id:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected 1 open order but saw {len(open_orders)} open orders")
    if round(open_orders[0].qty, 3) != round(size*0.75, 3):
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected open_orders[0].qty = {round(size*0.75, 3)} but saw {round(open_orders[0].qty, 3)}")
    print('success')

    taker_size = size*0.25
    print(f"taking {size} @ {price}...", end='', flush=True)
    taker.place_order(ctx, product, Side.BID, taker_size, price)
    print('success')
    
    print('waiting a few seconds for the crank...', end='', flush=True)
    await asyncio.sleep(CRANK_SLEEP_SECONDS)
    print('done waiting')
    print('checking open orders...', end='', flush=True)
    open_orders = list(maker.open_orders(ctx))
    if len(open_orders) != 1:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected 1 open order but saw {len(open_orders)} open orders")
    if open_orders[0].client_order_id != client_order_id:
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected 1 open order but saw {len(open_orders)} open orders")
    if round(open_orders[0].qty, 3) != round(size*0.5, 3):
        print("printing open orders")
        for o in open_orders:
            print(o)
        raise RuntimeError(f"expected open_orders[0].qty = {round(size*0.5, 3)} but saw {round(open_orders[0].qty, 3)}")
    print('success')


if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.codegen.dex.types import InitializeMarketProductGroupParams
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from solders.pubkey import Pubkey
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(initialize_market_product_group)]: DON'T MODIFY
@dataclass
class InitializeMarketProductGroupIx:
    program_id: Pubkey

    # account metas
    authority: AccountMeta
    market_product_group: AccountMeta
    market_product_group_vault: AccountMeta
    vault_mint: AccountMeta
    fee_collector: AccountMeta
    fee_model_program: AccountMeta
    fee_model_configuration_acct: AccountMeta
    risk_model_configuration_acct: AccountMeta
    risk_engine_program: AccountMeta
    sysvar_rent: AccountMeta
    system_program: AccountMeta
    token_program: AccountMeta
    fee_output_register: AccountMeta
    risk_output_register: AccountMeta
    staking_fee_collector: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    params: InitializeMarketProductGroupParams

    def to_instruction(self):
        keys = []
        keys.append(self.authority)
        keys.append(self.market_product_group)
        keys.append(self.market_product_group_vault)
        keys.append(self.vault_mint)
        keys.append(self.fee_collector)
        keys.append(self.fee_model_program)
        keys.append(self.fee_model_configuration_acct)
        keys.append(self.risk_model_configuration_acct)
        keys.append(self.risk_engine_program)
        keys.append(self.sysvar_rent)
        keys.append(self.system_program)
        keys.append(self.token_program)
        keys.append(self.fee_output_register)
        keys.append(self.risk_output_register)
        keys.append(self.staking_fee_collector)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.INITIALIZE_MARKET_PRODUCT_GROUP))
        buffer.write(BYTES_CATALOG.pack(InitializeMarketProductGroupParams, self.params))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(initialize_market_product_group)]: DON'T MODIFY
def initialize_market_product_group(
    authority: Union[str, Pubkey, AccountMeta],
    market_product_group: Union[str, Pubkey, AccountMeta],
    market_product_group_vault: Union[str, Pubkey, AccountMeta],
    vault_mint: Union[str, Pubkey, AccountMeta],
    fee_collector: Union[str, Pubkey, AccountMeta],
    fee_model_program: Union[str, Pubkey, AccountMeta],
    fee_model_configuration_acct: Union[str, Pubkey, AccountMeta],
    risk_model_configuration_acct: Union[str, Pubkey, AccountMeta],
    risk_engine_program: Union[str, Pubkey, AccountMeta],
    fee_output_register: Union[str, Pubkey, AccountMeta],
    risk_output_register: Union[str, Pubkey, AccountMeta],
    staking_fee_collector: Union[str, Pubkey, AccountMeta],
    params: InitializeMarketProductGroupParams,
    sysvar_rent: Union[str, Pubkey, AccountMeta] = Pubkey.from_string("SysvarRent111111111111111111111111111111111"),
    system_program: Union[str, Pubkey, AccountMeta] = Pubkey.from_string("11111111111111111111111111111111"),
    token_program: Union[str, Pubkey, AccountMeta] = Pubkey.from_string("TokenkegQfeZyiNwAJbNbGKPFXCWuBvf9Ss623VQ5DA"),
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL")

    if isinstance(authority, (str, Pubkey)):
        authority = to_account_meta(
            authority,
            is_signer=True,
            is_writable=False,
        )
    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(market_product_group_vault, (str, Pubkey)):
        market_product_group_vault = to_account_meta(
            market_product_group_vault,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(vault_mint, (str, Pubkey)):
        vault_mint = to_account_meta(
            vault_mint,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_collector, (str, Pubkey)):
        fee_collector = to_account_meta(
            fee_collector,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_program, (str, Pubkey)):
        fee_model_program = to_account_meta(
            fee_model_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_configuration_acct, (str, Pubkey)):
        fee_model_configuration_acct = to_account_meta(
            fee_model_configuration_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_model_configuration_acct, (str, Pubkey)):
        risk_model_configuration_acct = to_account_meta(
            risk_model_configuration_acct,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_engine_program, (str, Pubkey)):
        risk_engine_program = to_account_meta(
            risk_engine_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(sysvar_rent, (str, Pubkey)):
        sysvar_rent = to_account_meta(
            sysvar_rent,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(system_program, (str, Pubkey)):
        system_program = to_account_meta(
            system_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(token_program, (str, Pubkey)):
        token_program = to_account_meta(
            token_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_output_register, (str, Pubkey)):
        fee_output_register = to_account_meta(
            fee_output_register,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_output_register, (str, Pubkey)):
        risk_output_register = to_account_meta(
            risk_output_register,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(staking_fee_collector, (str, Pubkey)):
        staking_fee_collector = to_account_meta(
            staking_fee_collector,
            is_signer=False,
            is_writable=False,
        )

    return InitializeMarketProductGroupIx(
        program_id=program_id,
        authority=authority,
        market_product_group=market_product_group,
        market_product_group_vault=market_product_group_vault,
        vault_mint=vault_mint,
        fee_collector=fee_collector,
        fee_model_program=fee_model_program,
        fee_model_configuration_acct=fee_model_configuration_acct,
        risk_model_configuration_acct=risk_model_configuration_acct,
        risk_engine_program=risk_engine_program,
        sysvar_rent=sysvar_rent,
        system_program=system_program,
        token_program=token_program,
        fee_output_register=fee_output_register,
        risk_output_register=risk_output_register,
        staking_fee_collector=staking_fee_collector,
        remaining_accounts=remaining_accounts,
        params=params,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.dex.types.fractional import Fractional
from dexteritysdk.codegen.dex.types.product_metadata import ProductMetadata
from dexteritysdk.codegen.dex.types.product_status import ProductStatus
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    FixedLenArray,
    U64,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(Outright)]: DON'T MODIFY
@pod
class Outright:
    metadata: "ProductMetadata"
    num_risk_state_accounts: Usize
    product_status: "ProductStatus"
    dust: "Fractional"
    cum_funding_per_share: "Fractional"
    cum_social_loss_per_share: "Fractional"
    open_long_interest: "Fractional"
    open_short_interest: "Fractional"
    mark_price_qualifying_cum_value: "Fractional"
    mark_price_max_qualifying_width: "Fractional"
    padding: FixedLenArray[U64, 10]
    # LOCK-END

    def is_expired(self) -> bool:
        return self.product_status == ProductStatus.EXPIRED

    def is_active(self) -> bool:
        return self.product_status != ProductStatus.UNINITIALIZED

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.codegen.dex.types.order_info import OrderInfo
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.pubkey import Pubkey
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(validate_account_health)]: DON'T MODIFY
@dataclass
class ValidateAccountHealthIx:
    program_id: Pubkey

    # account metas
    market_product_group: AccountMeta
    trader_risk_group: AccountMeta
    risk_output_register: AccountMeta
    variance_cache: AccountMeta
    risk_model_configuration: AccountMeta
    risk_signer: AccountMeta
    covariance_metadata: AccountMeta
    correlation_matrix: AccountMeta
    mark_prices: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    params: OrderInfo

    def to_instruction(self):
        keys = []
        keys.append(self.market_product_group)
        keys.append(self.trader_risk_group)
        keys.append(self.risk_output_register)
        keys.append(self.variance_cache)
        keys.append(self.risk_model_configuration)
        keys.append(self.risk_signer)
        keys.append(self.covariance_metadata)
        keys.append(self.correlation_matrix)
        keys.append(self.mark_prices)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.VALIDATE_ACCOUNT_HEALTH))
        buffer.write(BYTES_CATALOG.pack(OrderInfo, self.params))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(validate_account_health)]: DON'T MODIFY
def validate_account_health(
    market_product_group: Union[str, Pubkey, AccountMeta],
    trader_risk_group: Union[str, Pubkey, AccountMeta],
    risk_output_register: Union[str, Pubkey, AccountMeta],
    variance_cache: Union[str, Pubkey, AccountMeta],
    risk_model_configuration: Union[str, Pubkey, AccountMeta],
    risk_signer: Union[str, Pubkey, AccountMeta],
    covariance_metadata: Union[str, Pubkey, AccountMeta],
    correlation_matrix: Union[str, Pubkey, AccountMeta],
    mark_prices: Union[str, Pubkey, AccountMeta],
    params: OrderInfo,
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("92wdgEqyiDKrcbFHoBTg8HxMj932xweRCKaciGSW3uMr")

    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(trader_risk_group, (str, Pubkey)):
        trader_risk_group = to_account_meta(
            trader_risk_group,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_output_register, (str, Pubkey)):
        risk_output_register = to_account_meta(
            risk_output_register,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(variance_cache, (str, Pubkey)):
        variance_cache = to_account_meta(
            variance_cache,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_model_configuration, (str, Pubkey)):
        risk_model_configuration = to_account_meta(
            risk_model_configuration,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(risk_signer, (str, Pubkey)):
        risk_signer = to_account_meta(
            risk_signer,
            is_signer=True,
            is_writable=False,
        )
    if isinstance(covariance_metadata, (str, Pubkey)):
        covariance_metadata = to_account_meta(
            covariance_metadata,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(correlation_matrix, (str, Pubkey)):
        correlation_matrix = to_account_meta(
            correlation_matrix,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(mark_prices, (str, Pubkey)):
        mark_prices = to_account_meta(
            mark_prices,
            is_signer=False,
            is_writable=True,
        )

    return ValidateAccountHealthIx(
        program_id=program_id,
        market_product_group=market_product_group,
        trader_risk_group=trader_risk_group,
        risk_output_register=risk_output_register,
        variance_cache=variance_cache,
        risk_model_configuration=risk_model_configuration,
        risk_signer=risk_signer,
        covariance_metadata=covariance_metadata,
        correlation_matrix=correlation_matrix,
        mark_prices=mark_prices,
        remaining_accounts=remaining_accounts,
        params=params,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.codegen.instruments.types.oracle_type import OracleType
from dexteritysdk.codegen.risk_engine.types.mark_price import MarkPrice
from podite import (
    FixedLenArray,
    U8,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(MarkPricesArray)]: DON'T MODIFY
@pod
class MarkPricesArray:
    bump: U8
    is_hardcoded_oracle: bool
    hardcoded_oracle: Pubkey
    _padding0: FixedLenArray[U8, 6]
    hardcoded_oracle_type: OracleType
    _padding1: FixedLenArray[U8, 7]
    array: FixedLenArray[MarkPrice, 64]
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from __future__ import annotations
from enum import IntEnum
from construct import Bytes
from construct import Switch
from borsh_construct import CStruct, U8, U64, U128, Bool

CALLBACK = CStruct(
    "account" / Bytes(32),
    "open_orders_idx" / U64,
    "order_nonce" / U128,
    "client_order_id" / U64,
)

EVENT_QUEUE_HEADER_LAYOUT = CStruct(
    "tag" / U8,
    "head" / U64,
    "count" / U64,
    "event_size" / U64,
    "seq_num" / U64,
)

# Option[OrderSummary]
REGISTER = CStruct(
    "register_present" / Bool,
    # Option[U128]
    "posted_order_id_present" / Bool,
    "posted_order_id" / U128,
    "total_base_qty" / U64,
    "total_quote_qty" / U64,
    "total_base_qty_posted" / U64,
)


class EventType(IntEnum):
    FILL = 0
    OUT = 1


class Side(IntEnum):
    BID = 0
    ASK = 1


class CompletedReason(IntEnum):
    CANCELLED = 0
    FILLED = 1
    BOOTED = 2
    SELF_TRADE_CANCEL = 3
    POST_NOT_ALLOWED = 4
    MATCH_LIMIT_EXHAUSTED = 5
    POST_ONLY = 6


# events
FILL_EVENT = CStruct(
    "taker_side" / U8,
    "maker_order_id" / U128,
    "quote_size" / U64,
    "base_size" / U64,
    "maker_callback" / CALLBACK,
    "taker_callback" / CALLBACK,
)

OUT_EVENT = CStruct(
    "side" / U8,
    "order_id" / U128,
    "base_size" / U64,
    "delete" / Bool,
    "callback" / CALLBACK,
    "reason" / U8,
)

EVENT_NODE_LAYOUT = CStruct(
    "tag" / U8,
    "data"
    / Switch(
        lambda this: this.tag,
        {
            EventType.FILL: FILL_EVENT,
            EventType.OUT: OUT_EVENT,
        },
    ),
)



# LOCK-BEGIN[imports]: DON'T MODIFY
from podite import (
    U8,
    pod,
)

# LOCK-END


# LOCK-BEGIN[class(UpdateMarkPricesParams)]: DON'T MODIFY
@pod
class UpdateMarkPricesParams:
    num_products: U8
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



from enum import IntEnum
from dataclasses import field
from typing import Union, List

from solders.pubkey import Pubkey

from dexteritysdk.utils import Usize
from podite import (
    U8,
    U64,
    U128,
    Option,
    pod,
    Enum,
    BYTES_CATALOG
)

from .base import AccountTag, Side


@pod
class EventQueueHeader:
    tag: AccountTag
    head: U64
    count: U64
    event_size: U64
    seq_num: U64
    unknown: Usize


@pod
class OrderSummary:
    posted_order_id: Option[U128]
    total_base_qty: U64
    total_quote_qty: U64
    total_base_qty_posted: U64


class EventKind(Enum[U8]):
    FILL = 0
    OUT = 1


@pod
class Callback:
    user_account: Pubkey


@pod
class FillEventData:
    taker_side: Side
    maker_order_id: U128
    quote_size: U64
    base_size: U64
    maker_callback_info: Callback
    taker_callback_info: Callback


@pod
class OutEventData:
    side: Side
    order_id: U128
    base_size: U64
    delete: bool
    callback_info: Callback


@pod
class Event:
    kind: EventKind
    event_data: Union[FillEventData, OutEventData]

    @classmethod
    def from_bytes_partial(cls, buffer):
        kind, buffer = BYTES_CATALOG.deserialize_to_type(EventKind, buffer)
        if kind == EventKind.FILL:
            event_data, _ = BYTES_CATALOG.deserialize_to_type(FillEventData, buffer)
        else:
            event_data, _ = BYTES_CATALOG.deserialize_to_type(OutEventData, buffer)

        max_size = max(
            BYTES_CATALOG.calcmaxsize_for_type(FillEventData),
            BYTES_CATALOG.calcmaxsize_for_type(OutEventData),
        )

        return Event(kind, event_data), buffer[max_size:]

    @classmethod
    def to_bytes_io(cls, obj, buffer):
        raise NotImplementedError

    def get_user_accounts(self) -> List[Pubkey]:
        if isinstance(self.event_data, FillEventData):
            return [
                self.event_data.maker_callback_info.user_account,
                self.event_data.taker_callback_info.user_account,
            ]
        else:
            return [
                self.event_data.callback_info.user_account,
            ]


@pod
class EventQueue:
    header: EventQueueHeader
    register: Option[OrderSummary]
    remaining_buffer: bytes = field(repr=False)

    def __getitem__(self, idx) -> Event:
        if idx >= self.header.count:
            raise ValueError("Index out of bound")

        head = self.header.head
        size = self.header.event_size
        buf_len = len(self.remaining_buffer)

        start = (head + idx * size) % buf_len
        end = start + size
        chunk = self.remaining_buffer[start:end]
        return Event.from_bytes_partial(chunk)[0]

    @classmethod
    def from_bytes_partial(cls, buffer):
        header, buffer = BYTES_CATALOG.deserialize_to_type(EventQueueHeader, buffer)
        register, buffer = BYTES_CATALOG.deserialize_to_type(
            Option[OrderSummary], buffer
        )

        return EventQueue(header, register, buffer), b""

    @classmethod
    def to_bytes_io(cls, obj, buffer):
        raise NotImplementedError



import base64
import json
from contextlib import contextmanager
from functools import wraps
from hashlib import sha256
from time import time, sleep
from typing import Optional, Dict
from typing import Union, Tuple, Iterable

import base58
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc.api import Client
from solana.rpc.commitment import Confirmed, Finalized
from solana.rpc.core import RPCException, TransactionExpiredBlockheightExceededError, UnconfirmedTxError
from solana.rpc.types import TxOpts
from solders.address_lookup_table_account import AddressLookupTableAccount
from solders.instruction import AccountMeta, Instruction
from solders.message import MessageV0
from solders.pubkey import Pubkey
from solders.signature import Signature
from solders.transaction import VersionedTransaction
from solana.transaction import Transaction
from solders.transaction_status import ParsedAccount

DEX_LOG_PREFIX = "Program log: dex-log "
DEX_ORDER_SUMMARY_PREFIX = DEX_LOG_PREFIX + "new-order:order-summary "
ERROR_PREFIX = "Program log: Error: "

def calc_rent(space, client=None):
    if client is None:
        client = Context.get_global_client()
    return client.get_minimum_balance_for_rent_exemption(space).value


class Context:
    client: Optional[Client] = None
    parser: Optional["AccountParser"] = None
    signers: Dict[bytes, Tuple[Keypair, str]] = {}
    fee_payer: Optional[Keypair] = None
    raise_on_error = False
    trader_nonce: int = 0

    @staticmethod
    def init_globals(
            fee_payer: Keypair,
            client: Client,
            signers: Iterable[Tuple[Keypair, str]],
            parser: Optional["AccountParser"] = None,
            raise_on_error=False,
    ):
        Context.fee_payer = fee_payer
        Context.client = client
        Context.parser = parser
        Context.signers = {}
        Context.raise_on_error = raise_on_error
        Context.add_signers(*signers)

    @staticmethod
    def get_global_client():
        return Context.client

    @staticmethod
    def set_global_client(client):
        Context.client = client

    @staticmethod
    def get_global_parser():
        return Context.parser

    @staticmethod
    def set_global_parser(parser):
        Context.parser = parser

    @staticmethod
    def get_signers():
        return Context.signers

    # todo: rename to add signers
    @staticmethod
    def add_signers(*signers: Tuple[Keypair, str], verify=True):
        for (signer, name) in signers:
            if not isinstance(signer, Keypair) or not isinstance(name, str):
                raise ValueError(f"signers must be a list iterable of (Keypair, str) tuples. Found: {signer, name}")
            if bytes(signer.pubkey()) not in Context.signers:
                Context.signers[bytes(signer.pubkey())] = (signer, name)
        if verify:
            names = set()
            for (_, name) in Context.signers.values():
                if name in names:
                    raise ValueError("Each signer name must be unique")
                names.add(name)

    @staticmethod
    def get_global_fee_payer():
        return Context.fee_payer

    @staticmethod
    def set_global_fee_payer(fee_payer: Keypair):
        Context.fee_payer = fee_payer

    @staticmethod
    def get_raise_on_error():
        return Context.raise_on_error

    @staticmethod
    def set_raise_on_error(raise_on_error: bool):
        Context.raise_on_error = raise_on_error


@contextmanager
def global_fee_payer(fee_payer):
    old_fee_payer = Context.get_global_fee_payer()
    Context.set_global_fee_payer(fee_payer)
    yield
    Context.set_global_fee_payer(old_fee_payer)


class AccountParser:
    _parsers: Dict[bytes, callable]  # key: program_id

    def __init__(self):
        self._parsers = dict()

    def register_parser(self, program_id, parser):
        self._parsers[bytes(program_id)] = parser

    def register_parser_from_account_enum(self, program_id: Pubkey, accounts_enum):
        def parser(info):
            return accounts_enum.from_bytes(info).field

        self.register_parser(program_id, parser)

    def parse(self, owner, data):
        try:
            parser = self._parsers[bytes(owner)]
        except Exception as e:
            raise ValueError(f"Failed to find parser corresponding to account owner. Owner={owner}",
                             [str(base58.b58encode(p)) for p in self._parsers.keys()])
        return parser(data)


class TransactionDetails:
    def __init__(self, content, cluster="devnet"):
        self.content = content
        self.cluster = cluster

    def is_valid_tx(self):
        return self.content.value is not None

    @property
    def account_keys(self):
        accounts = self.content.value.transaction.transaction.message.account_keys
        account_keys = []
        for account in accounts:
            if isinstance(account, ParsedAccount):
                account_keys.append(account.pubkey())
            else:
                account_keys.append(account)
        return account_keys

    @property
    def signatures(self):
        return self.content.value.transaction.transaction.signatures

    @property
    def tx_string(self):
        return self.signatures[0]

    @property
    def log_messages(self):
        return self.content.value.transaction.meta.log_messages

    @property
    def emitted_logs(self):
        result = dict()
        for msg in self.log_messages:
            if msg.startswith(DEX_LOG_PREFIX):
                key, val = msg[len(DEX_LOG_PREFIX):].split(" ")
                result[key] = base64.b64decode(val)

        return result

    def emitted_dex_order_summaries(self):
        for msg in self.log_messages:
            if msg.startswith(DEX_ORDER_SUMMARY_PREFIX):
                yield base64.b64decode(msg[len(DEX_ORDER_SUMMARY_PREFIX):])

    @property
    def error_from_log(self):
        errs = []
        for msg in self.log_messages:
            if msg.startswith(ERROR_PREFIX):
                errs.append(msg[len(ERROR_PREFIX):])
        if len(errs) == 0:
            return None
        return errs[-1]

    @property
    def error(self):
        return self.content.value.transaction.meta.err

    def __str__(self) -> str:
        return f"TransactionDetails({self.tx_string})"

    def __repr__(self) -> str:
        return str(self)


class AccountDetails:
    def __init__(self, pubkey, content, cluster="devnet"):
        self.pubkey = pubkey
        self.content = content
        self.cluster = cluster

    def __str__(self) -> str:
        return f"AccountDetails({self.pubkey})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def slot(self):
        return self.content.context.slot

    @property
    def data(self):
        if not self.content.value:
            return None

        return self.content.value.data

    @property
    def data_obj(self):
        parser = Context.get_global_parser()
        return parser.parse(self.content.value.owner, self.content.value.data)


class MultipleAccountDetails:
    def __init__(self, pubkeys, content, cluster="devnet"):
        self.pubkeys = pubkeys
        self.content = content
        self.cluster = cluster

    def __str__(self) -> str:
        return f"MultipleAccountDetails({self.pubkeys})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def slot(self):
        return self.content.context.slot

    @property
    def data(self):
        if not self.content.value:
            return None

        return [acc_value.data for acc_value in self.content.value]

    @property
    def data_objs(self):
        parser = Context.get_global_parser()
        return [parser.parse(acc_value.owner, acc_value.data) for acc_value in self.content.value]


def fetch_transaction_details(sig, client=None):
    if client is None:
        client = Context.get_global_client()

    content = client.get_transaction(
        sig,
        encoding="json",
        commitment=Confirmed,
        max_supported_transaction_version=0,
    )
    return TransactionDetails(content)


def fetch_account_details(addr, client=None):
    if client is None:
        client = Context.get_global_client()

    content = client.get_account_info(addr, commitment=Confirmed)
    return AccountDetails(addr, content)


def fetch_multiple_account_details(addresses, client=None):
    if client is None:
        client = Context.get_global_client()

    content = client.get_multiple_accounts(addresses, commitment=Confirmed)
    return MultipleAccountDetails(addresses, content)


# only supports accounts
def explore_many(addresses):
    pubkeys = []
    for addr in addresses:
        if isinstance(addr, str):
            if len(base58.b58decode(addr)) == 64:
                raise ValueError()
            else:
                pubkeys.append(Pubkey.from_string(addr))
        elif isinstance(addr, Pubkey):
            pubkeys.append(addr)
        elif isinstance(addr, Signature):
            raise ValueError()
        else:
            raise ValueError()

    return fetch_multiple_account_details(pubkeys)


def explore(addr):
    if isinstance(addr, str):
        if len(base58.b58decode(addr)) == 64:
            kind = "tx"
            addr = Signature.from_string(addr)
        else:
            kind = "acc"
            addr = Pubkey.from_string(addr)
    elif isinstance(addr, Pubkey):
        kind = "acc"
    elif isinstance(addr, Signature):
        kind = "tx"
    else:
        raise ValueError()

    if kind == "tx":
        return fetch_transaction_details(addr)
    else:
        return fetch_account_details(addr)


def send_instructions(
        *ixs: Instruction,
        **kwargs
) -> Union[TransactionDetails, AccountDetails]:
    return send_transaction(Transaction().add(*ixs), **kwargs)


def legacy_to_v0(
        tx,
        signers,
        recent_blockhash,
        address_lookup_table_accounts,
) -> VersionedTransaction:
    msg = MessageV0.try_compile(
        payer=tx.fee_payer,
        instructions=tx.instructions,
        address_lookup_table_accounts=address_lookup_table_accounts,
        recent_blockhash=recent_blockhash
    )
    return VersionedTransaction(msg, signers)

def send_transaction(
        tx,
        *signers: Keypair,
        recent_blockhash=None,
        client=None,
        raise_on_error=None,
        confirm_tx_timeout=120,
        fetch_tx_timeout=120,
        address_lookup_table_accounts=None, # otherwise, list of AddressLookupTableAccount
        opts=None,
) -> Union[TransactionDetails, AccountDetails]:
    if fee_payer := Context.get_global_fee_payer():
        if isinstance(fee_payer, Keypair):
            tx = Transaction(fee_payer=fee_payer.pubkey()).add(tx)
        elif isinstance(fee_payer, Pubkey):
            tx = Transaction(fee_payer=fee_payer).add(tx)
        else:
            raise ValueError("fee_payer must be a Keypair or a Pubkey, but it's a {type(fee_payer)}")

    raise_on_error = raise_on_error if raise_on_error is not None else Context.get_raise_on_error()

    if len(signers) == 0:
        signers = Context.get_signers()
    else:
        signers = {bytes(signer.pubkey()): (signer, f"arg  {i}") for i, signer in enumerate(signers)}

    if client is None:
        client = Context.get_global_client()

    # filtering private keys to only contain the relevant ones
    # otherwise, there will be a problem with fee_payer
    signers_pubkeys = []
    if tx.fee_payer:
        signers_pubkeys.append(tx.fee_payer)
    for ix in tx.instructions:
        for i, meta in enumerate(ix.accounts):
            if not isinstance(meta, AccountMeta):
                print(f'{i} is {meta}')
            if meta.is_signer and meta.pubkey not in signers_pubkeys:
                signers_pubkeys.append(meta.pubkey)
    signer_keypairs = []
    for pk in signers_pubkeys:
        if bytes(pk) not in signers:
            names = [(name, str(base58.b58encode(p))) for p, (_, name) in signers.items()]
            raise ValueError(f"Required signer Pubkey not in list of Keypairs. Have {names}, want: {pk}")
        signer_keypairs.append(signers[bytes(pk)][0])

    blockhash_resp = client.get_latest_blockhash(Finalized)
    if not recent_blockhash:
        recent_blockhash = client.parse_recent_blockhash(blockhash_resp)
    last_valid_block_height = blockhash_resp.value.last_valid_block_height

    opts = opts if opts is not None else TxOpts(skip_preflight=True, skip_confirmation=True, preflight_commitment=Confirmed,
                    last_valid_block_height=last_valid_block_height)

    # print(f"address_lookup_table_accounts: {address_lookup_table_accounts}")
    # send tx without confirmation, we'll confirm manually right after with better error handling
    if address_lookup_table_accounts is not None:
        # print("converting legacy tx to v0 using ALTs")
        tx = legacy_to_v0(tx, signer_keypairs, recent_blockhash, address_lookup_table_accounts)
        result = client.send_transaction(
            tx,
            opts=opts,
        )
    else:
        result = client.send_transaction(
            tx,
            *signer_keypairs,
            opts=opts,
            recent_blockhash=recent_blockhash,
        )

    # print(f"sent tx! {result.value}")

    # try to confirm tx
    timeout = time() + confirm_tx_timeout
    while time() < timeout:
        try:
            client.confirm_transaction(
                tx_sig=result.value,
                commitment=Confirmed,
                last_valid_block_height=last_valid_block_height,
            )
            break
        except TransactionExpiredBlockheightExceededError:
            raise
        except Exception as e:
            print(f"continuing on, but failed to confirm tx with error {e}")

        blockhash_resp = client.get_latest_blockhash(Finalized)
        last_valid_block_height = blockhash_resp.value.last_valid_block_height
        sleep(0.5)
    else:
        raise UnconfirmedTxError(f"Unable to confirm transaction {result.value}")

    # try to fetch transaction data (logs etc.)
    exc = None
    timeout = time() + fetch_tx_timeout
    while time() < timeout:
        if exc:
            raise exc

        try:
            transaction_details = explore(result.value)
            if transaction_details.is_valid_tx():
                if transaction_details.error and raise_on_error:
                    try:
                        if hasattr(transaction_details.error, 'to_json'):
                            err_str = transaction_details.error.__class__.__name__ + " " + transaction_details.error.to_json()
                        else:
                            err_str = json.dumps(transaction_details.error)
                    except:
                        err_str = transaction_details.error.__class__.__name__
                    log_str = "\n".join(transaction_details.log_messages)
                    exc = ValueError(f"Transaction returned error:\n{err_str}, \nLog messages:\n{log_str}")
                    continue

                return transaction_details
            else:
                raise ValueError(f"found invalid tx. transaction_details: {transaction_details}")
        except Exception as e:
            print(f"continuing on, but failed to fetch tx details with error {e}")

        sleep(0.5)

    raise RPCException(f"Failed to fetch confirmed tx {result.value}")


def actionify(func=None, /, post_process=lambda resp: (None, resp), raise_error=False):
    assert not raise_error, "Raise_error is not implemented"

    def _actionify(make):
        @wraps(make)
        def send(*args, **kwargs):
            tx = make(*args, **kwargs)
            if tx is None:
                return post_process(None)
            if isinstance(tx, Instruction):
                tx = Transaction().add(tx)

            opts = TxOpts(
                skip_preflight=True,
                skip_confirmation=False,
                preflight_commitment=Confirmed,
            )
            response = send_transaction(
                tx,
                opts=opts,
            )
            return post_process(response)

        send.make = make
        return send

    if func is None:
        return _actionify
    return _actionify(func)


def sighash(ix_name: str) -> bytes:
    """Not technically sighash, since we don't include the arguments.
    (Because Rust doesn't allow function overloading.)
    Args:
        ix_name: The instruction name.
    Returns:
        The sighash bytes.
    """
    formatted_str = f"global:{ix_name}"
    return sha256(formatted_str.encode()).digest()[:8]


def sighash_int(ix_name: str) -> int:
    return int.from_bytes(sighash(ix_name), byteorder="little")

def get_address_lookup_table_account(pubkey) -> AddressLookupTableAccount:
    alta_account = fetch_account_details(pubkey)
    # Idiomatic, but I think this method was never implemented in solders:
    # address_lookup_table_account = AddressLookupTableAccount.from_bytes(alta_account.data)
    # So instead, we must manually construct the ALT
    addresses = []
    alta_data = alta_account.data
    for offset in range(56, len(alta_data), 32):
        pk = Pubkey.from_bytes(alta_data[offset:offset+32])
        addresses.append(pk)
    return AddressLookupTableAccount(
        key=pubkey,
        addresses=addresses
    )



import podite
from podite import BYTES_CATALOG, JSON_CATALOG, pod
from podite._utils import get_concrete_type, get_calling_module, _GetitemToCall


def _open_orders_array(name, type_):
    module = get_calling_module()

    max_length = 1024

    @pod(dataclass_fn=None)
    class _ArrayPod:
        length_relative_offset = -6148

        @classmethod
        def _is_static(cls) -> bool:
            return BYTES_CATALOG.is_static(get_concrete_type(module, type_))

        @classmethod
        def _calc_size(cls, obj, **kwargs):
            ty = get_concrete_type(module, type_)
            body_size = sum(
                (BYTES_CATALOG.calc_size(ty, elem, **kwargs) for elem in obj)
            )
            return body_size

        @classmethod
        def _calc_max_size(cls):
            return (
                BYTES_CATALOG.calc_max_size(get_concrete_type(module, type_)) * max_length
            )

        @classmethod
        def _from_bytes_partial(cls, buffer, **kwargs):
            pos = buffer.tell()
            buffer.seek(length_relative_offset - max_length, 1)
            max_open_orders = BYTES_CATALOG.unpack_partial(podite.U16, buffer)
            buffer.seek(pos)

            result = []
            for _ in range(max_open_orders):
                value = BYTES_CATALOG.unpack_partial(
                    get_concrete_type(module, type_), buffer, **kwargs
                )
                result.append(value)

            return result

        @classmethod
        def _to_bytes_partial(cls, buffer, obj, **kwargs):
            if len(obj) > max_length:
                raise RuntimeError("actual_length > max_length")
            for elem in obj:
                BYTES_CATALOG.pack_partial(
                    get_concrete_type(module, type_), buffer, elem
                )

        @classmethod
        def _to_dict(cls, obj):
            return [JSON_CATALOG.pack(get_concrete_type(module, type_), e) for e in obj]

        @classmethod
        def _from_dict(cls, raw):
            return [
                JSON_CATALOG.unpack(get_concrete_type(module, type_), e) for e in raw
            ]

    _ArrayPod.__name__ = f"{name}[{type_}, {_ArrayPod.length_relative_offset}]"
    _ArrayPod.__qualname__ = _ArrayPod.__name__

    return _ArrayPod


OpenOrdersArray = _GetitemToCall("FixedLenArray", _open_orders_array)



# LOCK-BEGIN[imports]: DON'T MODIFY
from .instruction_tag import InstructionTag
from dataclasses import dataclass
from dexteritysdk.codegen.dex.types import ConsumeOrderbookEventsParams
from dexteritysdk.solmate.utils import to_account_meta
from io import BytesIO
from podite import BYTES_CATALOG
from solders.instruction import (
    AccountMeta,
    Instruction,
)
from solders.pubkey import Pubkey
from typing import (
    List,
    Optional,
    Union,
)

# LOCK-END


# LOCK-BEGIN[ix_cls(consume_orderbook_events)]: DON'T MODIFY
@dataclass
class ConsumeOrderbookEventsIx:
    program_id: Pubkey

    # account metas
    aaob_program: AccountMeta
    market_product_group: AccountMeta
    product: AccountMeta
    market_signer: AccountMeta
    orderbook: AccountMeta
    event_queue: AccountMeta
    reward_target: AccountMeta
    fee_model_program: AccountMeta
    fee_model_configuration_acct: AccountMeta
    fee_output_register: AccountMeta
    risk_and_fee_signer: AccountMeta
    remaining_accounts: Optional[List[AccountMeta]]

    # data fields
    params: ConsumeOrderbookEventsParams

    def to_instruction(self):
        keys = []
        keys.append(self.aaob_program)
        keys.append(self.market_product_group)
        keys.append(self.product)
        keys.append(self.market_signer)
        keys.append(self.orderbook)
        keys.append(self.event_queue)
        keys.append(self.reward_target)
        keys.append(self.fee_model_program)
        keys.append(self.fee_model_configuration_acct)
        keys.append(self.fee_output_register)
        keys.append(self.risk_and_fee_signer)
        if self.remaining_accounts is not None:
            keys.extend(self.remaining_accounts)

        buffer = BytesIO()
        buffer.write(InstructionTag.to_bytes(InstructionTag.CONSUME_ORDERBOOK_EVENTS))
        buffer.write(BYTES_CATALOG.pack(ConsumeOrderbookEventsParams, self.params))

        return Instruction(
            accounts=keys,
            program_id=self.program_id,
            data=buffer.getvalue(),
        )

# LOCK-END


# LOCK-BEGIN[ix_fn(consume_orderbook_events)]: DON'T MODIFY
def consume_orderbook_events(
    aaob_program: Union[str, Pubkey, AccountMeta],
    market_product_group: Union[str, Pubkey, AccountMeta],
    product: Union[str, Pubkey, AccountMeta],
    market_signer: Union[str, Pubkey, AccountMeta],
    orderbook: Union[str, Pubkey, AccountMeta],
    event_queue: Union[str, Pubkey, AccountMeta],
    reward_target: Union[str, Pubkey, AccountMeta],
    fee_model_program: Union[str, Pubkey, AccountMeta],
    fee_model_configuration_acct: Union[str, Pubkey, AccountMeta],
    fee_output_register: Union[str, Pubkey, AccountMeta],
    risk_and_fee_signer: Union[str, Pubkey, AccountMeta],
    params: ConsumeOrderbookEventsParams,
    remaining_accounts: Optional[List[AccountMeta]] = None,
    program_id: Optional[Pubkey] = None,
):
    if program_id is None:
        program_id = Pubkey.from_string("FUfpR31LmcP1VSbz5zDaM7nxnH55iBHkpwusgrnhaFjL")

    if isinstance(aaob_program, (str, Pubkey)):
        aaob_program = to_account_meta(
            aaob_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(market_product_group, (str, Pubkey)):
        market_product_group = to_account_meta(
            market_product_group,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(product, (str, Pubkey)):
        product = to_account_meta(
            product,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(market_signer, (str, Pubkey)):
        market_signer = to_account_meta(
            market_signer,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(orderbook, (str, Pubkey)):
        orderbook = to_account_meta(
            orderbook,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(event_queue, (str, Pubkey)):
        event_queue = to_account_meta(
            event_queue,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(reward_target, (str, Pubkey)):
        reward_target = to_account_meta(
            reward_target,
            is_signer=True,
            is_writable=True,
        )
    if isinstance(fee_model_program, (str, Pubkey)):
        fee_model_program = to_account_meta(
            fee_model_program,
            is_signer=False,
            is_writable=False,
        )
    if isinstance(fee_model_configuration_acct, (str, Pubkey)):
        fee_model_configuration_acct = to_account_meta(
            fee_model_configuration_acct,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(fee_output_register, (str, Pubkey)):
        fee_output_register = to_account_meta(
            fee_output_register,
            is_signer=False,
            is_writable=True,
        )
    if isinstance(risk_and_fee_signer, (str, Pubkey)):
        risk_and_fee_signer = to_account_meta(
            risk_and_fee_signer,
            is_signer=False,
            is_writable=False,
        )

    return ConsumeOrderbookEventsIx(
        program_id=program_id,
        aaob_program=aaob_program,
        market_product_group=market_product_group,
        product=product,
        market_signer=market_signer,
        orderbook=orderbook,
        event_queue=event_queue,
        reward_target=reward_target,
        fee_model_program=fee_model_program,
        fee_model_configuration_acct=fee_model_configuration_acct,
        fee_output_register=fee_output_register,
        risk_and_fee_signer=risk_and_fee_signer,
        remaining_accounts=remaining_accounts,
        params=params,
    ).to_instruction()

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
import dexteritysdk.codegen.dex.accounts as accounts
import dexteritysdk.codegen.dex.constants as constants
import dexteritysdk.codegen.dex.instructions as instructions
import dexteritysdk.codegen.dex.types as types

# LOCK-END



# LOCK-BEGIN[imports]: DON'T MODIFY
from dexteritysdk.solmate.dtypes import Usize
from podite import (
    I64,
    pod,
)
from solders.pubkey import Pubkey

# LOCK-END


# LOCK-BEGIN[class(Leg)]: DON'T MODIFY
@pod
class Leg:
    product_index: Usize
    product_key: Pubkey
    ratio: I64
    # LOCK-END

    @classmethod
    def to_bytes(cls, obj, **kwargs):
        return cls.pack(obj, converter="bytes", **kwargs)

    @classmethod
    def from_bytes(cls, raw, **kwargs):
        return cls.unpack(raw, converter="bytes", **kwargs)



import random
import string


def get_shared_random():
    return getattr(random, "_inst")


def random_hash(length=8, chars=string.ascii_uppercase + string.digits, seed=None):
    r = get_shared_random() if seed is None else random.Random(seed)
    return "".join(r.choice(chars) for _ in range(length))



import os
from dexteritysdk.local_env import *
import time
import sys
import asyncio
import base64
import json
import base64
import argparse
import websockets as ws

from solders.keypair import Keypair
from solders.pubkey import Pubkey

# dex
from dexteritysdk.dex.state import *
from dexteritysdk.dex.state import account_parser as dex_parser
import dexteritysdk.dex.actions as dex_act

# derivatives
from dexteritysdk.instruments.state import *
from dexteritysdk.instruments.state import account_parser as instr_parser
from dexteritysdk.dummy_oracle.state import *
from dexteritysdk.utils.aob.state import *
from dexteritysdk.utils.aob.state import account_parser as aob_parser
from dexteritysdk.utils.solana import (
    explore,
    AccountParser,
    Context
)
import dexteritysdk.program_ids as pids

# RPC
from solana.rpc.commitment import *
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc.api import Client

curr_directory = os.path.dirname(os.path.abspath(__file__))
cfg_path = os.path.join(curr_directory, "config.json")
if os.path.exists(cfg_path):
    with open(cfg_path) as f:
        cfg = json.load(f)
        mpg_pubkey = cfg["mpg_key"]
else:
    ap = argparse.ArgumentParser()
    ap.add_argument("mpg_pubkey")
    args = ap.parse_args()
    mpg_pubkey = args.mpg_pubkey

payer = Keypair(b"nima" * 8)
market_product_group_pubkey = Pubkey.from_string(mpg_pubkey)
reward_target = payer.pubkey
set_signers(payer)


def get_account_info(event):
    try:
        return [event.event_data.callback_info.user_account]
    except:
        return [
            event.event_data.maker_callback_info.user_account,
            event.event_data.taker_callback_info.user_account,
        ]


seq = 0


def get_rpc_payload(key):
    global seq
    payload = {
        "jsonrpc": "2.0",
        "id": seq,
        "method": "accountSubscribe",
        "params": [
            key,
            {"encoding": "base64", "commitment": "confirmed"},
        ],
    }
    seq += 1
    return payload


URL = "https://api.devnet.solana.com"
WS_URI = "wss://api.devnet.solana.com/"


type_dict = {}
event_queue_dict = {}
id_to_dict = {}

subscribed_products = []

parser = AccountParser()
parser.register_parser(pids.DEX_PROGRAM_ID, dex_parser)
parser.register_parser(pids.INSTRUMENTS_PROGRAM_ID, instr_parser)
parser.register_parser(pids.AOB_PROGRAM_ID, aob_parser)
set_global_parser(parser)
set_global_client(Client(URL))

market_product_group = explore(
    market_product_group_pubkey
).data_obj  # type: MarketProductGroup

print(mpg_pubkey)
print(market_product_group)

id_to_dict[seq] = {
    "type": "mpg",
}
mpg_payload = get_rpc_payload(str(market_product_group_pubkey))


async def register_events(market_product_group: MarketProductGroup, w):
    for product in (
        market_product_group.market_products + market_product_group.combo_products
    ):
        if product:
            print(product)
            product_pubkey = product.product_key
            if not product_pubkey in subscribed_products:
                orderbook_pubkey = product.orderbook
                id_to_dict[seq] = {
                    "type": "orderbook",
                    "product_pubkey": product_pubkey,
                    "orderbook_pubkey": orderbook_pubkey,
                }
                orderbook_payload = get_rpc_payload(str(orderbook_pubkey))
                orderbook = explore(orderbook_pubkey).data_obj  # type: MarketState
                event_pubkey = orderbook.event_queue
                await w.send(json.dumps(orderbook_payload))
                id_to_dict[seq] = {
                    "type": "event",
                    "event_pubkey": event_pubkey,
                    "product_pubkey": product_pubkey,
                    "orderbook_pubkey": orderbook_pubkey,
                    "orderbook_fee_budget": orderbook.fee_budget,
                }
                event_payload = get_rpc_payload(str(event_pubkey))
                await w.send(json.dumps(event_payload))
                subscribed_products.append(product_pubkey)

    print("all products loaded")


async def consume_events(payload, type_dict, client, n=20):
    event_data = base64.b64decode(payload["params"]["result"]["value"]["data"][0])
    events = EventQueue.from_bytes(event_data)
    product_pubkey = type_dict[payload["params"]["subscription"]]["product_pubkey"]
    orderbook_fee_budget = type_dict[payload["params"]["subscription"]][
        "orderbook_fee_budget"
    ]
    user_accounts = []
    for i in range(min(events.header.count, n)):
        user_accounts.extend(get_account_info(events[i]))

    unique_user_accounts = [
        Pubkey.from_string(item) for item in list(set([str(a) for a in user_accounts]))
    ]

    if events.header.count == 0:
        print("zero events")
        return

    capped_entries_consumed = min(n, events.header.count)
    reward = capped_entries_consumed * orderbook_fee_budget / events.header.count
    profit = reward - 5000
    print(f"profit after consuming this event: {profit}")
    if profit <= 0:
        print("It is not profitable to crank.")
        return

    pre_balance = client.get_balance(payer.pubkey())["result"]["value"]
    print("sending consume event ...")
    _, resp = dex_act.consume_orderbook_events(
        market_product_group=market_product_group_pubkey,
        product_key=product_pubkey,
        reward_target=reward_target,
        user_accounts=sorted(unique_user_accounts, key=lambda x: bytes(x)),
        max_iterations=n,
    )
    print("sent consume event!")
    print(resp.tx_string)

    post_balance = client.get_balance(payer.pubkey())["result"]["value"]
    print(f"Pre: {pre_balance}, Psot: {post_balance}")

    if post_balance < pre_balance:
        print("Lost money in cranking!")
        return


async def run(n=10):
    global market_product_group
    client = get_global_client()
    while True:
        async with ws.connect(WS_URI) as w:
            time.sleep(1)
            await w.send(json.dumps(mpg_payload))
            await register_events(market_product_group, w)
            while True:
                try:
                    acct = await (w.recv())
                except:
                    continue
                try:
                    payload = json.loads(acct)
                    if "params" in payload.keys():
                        if (
                            type_dict[payload["params"]["subscription"]]["type"]
                            == "mpg"
                        ):
                            print("market product group update")
                            market_product_group = parser.parse(
                                payload["params"]
                            )  # type: MarketProductGroup
                            await register_events(market_product_group, w)

                        elif (
                            type_dict[payload["params"]["subscription"]]["type"]
                            == "orderbook"
                        ):
                            print("orderbook update")
                            order_book: MarketState = parser.parse(payload["params"])
                            orderbook_pubkey = type_dict[
                                payload["params"]["subscription"]
                            ]["orderbook_pubkey"]
                            for item in type_dict:
                                try:
                                    if (
                                        type_dict[item]["orderbook_pubkey"]
                                        == orderbook_pubkey
                                    ):
                                        type_dict[item][
                                            "orderbook_fee_budget"
                                        ] = order_book.fee_budget
                                except:
                                    pass

                        elif (
                            type_dict[payload["params"]["subscription"]]["type"]
                            == "event"
                        ):
                            print("event update")
                            await consume_events(payload, type_dict, client, n)

                        else:
                            raise Exception("Can not find relevant payload!")

                    elif "result" in payload.keys():
                        try:
                            type_dict[payload["result"]] = id_to_dict[payload["id"]]
                        except:
                            raise Exception("Can not find relevant payload!")
                    else:
                        Exception("Can not find relevant payload!")
                except Exception as e:
                    print(e)
                    break


asyncio.run(run())
